#include <math.h>
#include "axis.h"

#define max(x,y) ((x)>(y) ? (x) : (y))
#define min(x,y) ((x)<(y) ? (x) : (y))


void get_order_of_magnitude (double val, double sc, double *man, double *pw)
{
    double b;
    double sx;

    b = (val != 0. ? log10 (fabs (val)) / sc : 0.);
    b = sc * floor (b);
    sx = (b != 0. ? pow (10., b) : 1.);
    *man = val / sx;
    *pw = b;

    return;
}
/********************************************************************
 * GetBins
 *
 *   Created: Tue May 04 21:15:45 1999        Author: Carlos
 *   Purpose: Finds out the binning for the scale of the control
 *            Wmin,Wmax       : range of the scale
 *            Nold            : wanted number of bins
 *            BinLow, BinHigh : Optimized low and hig bins for which
 *                              a label is to be drawn
 *            Nbins           : number of bins
 *            Bwidth          : bin width between ticks
 *
 ********************************************************************/
void getbins (float Wmin, float Wmax, int Nold, float *BinLow, float *BinHigh, int *Nbins, float *Bwidth)
{
    int lwid, kwid;
    int ntemp = 0;
    int jlog = 0;
    float siground = 0;
    float alb, awidth, sigfig, atest;

    float AL = min (Wmin, Wmax);
    float AH = max (Wmin, Wmax);
    if (AL == AH)
	AH = AL + 1;

    /*
     IF Nold  ==  -1 , program uses binwidth input from calling
      routine
    */
    if (Nold == -1 && (*Bwidth) > 0)
	goto L90;
    ntemp = max (Nold, 2);
    if (ntemp < 1)
	ntemp = 1;

    /* Get nominal bin width in exponential form */

 L20:
    awidth = (AH - AL) / ((float) ntemp);
    if (awidth > 1.e30)
	goto LOK;
    jlog = (int) (log10 (awidth));
    if (awidth <= 1)
	jlog--;
    sigfig = awidth * ((float) pow (10, -jlog));

    /* Round mantissa up to 1, 2, 2.5, 5, or 10 */

    if (sigfig <= 1)
	siground = 1;
    else if (sigfig <= 2)
	siground = 2;
    /*  else if (sigfig <= 2.5)  siground = 2.5; */
    else if (sigfig <= 5)
	siground = 5;
    else {
	siground = 1;
	jlog++;
    }

    (*Bwidth) = siground * ((float) pow (10, jlog));

    /*  Get new bounds from new width Bwidth */

 L90:
    alb = AL / (*Bwidth);
    lwid = (int) (alb);
    if (alb < 0)
	lwid--;
    (*BinLow) = siground * ((float) lwid) * ((float) pow (10, jlog));
    alb = AH / (*Bwidth) + 1.00001f;
    kwid = (int) alb;
    if (alb < 0)
	kwid--;
    (*BinHigh) = siground * ((float) kwid) * ((float) pow (10, jlog));
    (*Nbins) = kwid - lwid;
    if (Nold == -1)
	goto LOK;
    if (Nold <= 5) {
	/* Request for one bin is difficult case */
	if (Nold > 1 || (*Nbins) == 1)
	    goto LOK;
	(*Bwidth) = (*Bwidth) * 2;
	(*Nbins) = 1;
	goto LOK;
    }
    if (2 * (*Nbins) == Nold) {
	ntemp++;
	goto L20;
    }

 LOK:
    atest = (*Bwidth) * 0.0001f;
    if (fabs ((*BinLow) - Wmin) >= atest)
	(*BinLow) += (*Bwidth);
    (*Nbins)--;
    if (fabs ((*BinHigh) - Wmax) >= atest)
	(*BinHigh) -= (*Bwidth);
    (*Nbins)--;

}
