#include <stdio.h>
//#include <values.h>
#include "function.h"
#include "function-protected.h"


enum {
    RANGE_CHANGED,
    LAST_SIGNAL
};
enum {
    PROP_0,
    PROP_NAME,
    PROP_RANGE,
    PROP_NPTS,
    PROP_XMIN,
    PROP_XMAX,
    PROP_XSTEP,
    PROP_FCN
};

#define __swap_double(x,y) {                    \
        double tmp = x;                         \
        x = y;                                  \
        y = tmp;                                \
    }

static void       function_base_class_init     (HstFunctionClass *class);
static void       function_base_class_finalize (HstFunctionClass *class);
static void       function_class_init          (HstFunctionClass *klass);
static void       function_init                (HstFunction      *object,
                                                HstFunctionClass *klass);

static void       function_set_property   (GObject       *object,
                                           guint           property_id,
                                           const GValue   *value,
                                           GParamSpec     *pspec);
static void       function_get_property   (GObject       *object,
                                           guint           property_id,
                                           GValue         *value,
                                           GParamSpec     *pspec);
static void       function_dispose             (GObject        *object);
static void       function_finalize            (GObject        *object);
static void       function_array(HstFunction *f, Tuple *, double xmin, double xmax);
static int        function_resize(HstFunction *f,int width, int height);
static double     function_evaluate(HstFunction *f, double x);
static void       function_notify_range_changed(GObject *o, GParamSpec *arg, gpointer data);
static void       function_real_range_changed(HstFunction *f);
static gpointer    parent_class = NULL;
static guint       function_signals[LAST_SIGNAL] = { 0 };

/* --- functions --- */
GType hst_function_get_type (void) {
    static GType object_type = 0;

    if (!object_type) {
        static const GTypeInfo object_info = {
            sizeof (HstFunctionClass),
            (GBaseInitFunc) function_base_class_init,
            (GBaseFinalizeFunc) function_base_class_finalize,
            (GClassInitFunc) function_class_init,
            NULL,               /* class_finalize */
            NULL,               /* class_data */
            sizeof (HstFunction),
            16,         /* n_preallocs */
            (GInstanceInitFunc) function_init,
            NULL,               /* value_table */
        };

        object_type = g_type_register_static (G_TYPE_OBJECT, "HstFunction",
                                              &object_info, 0);
    }
    return object_type;
}

static void function_base_class_init (HstFunctionClass *class) {
    /* reset instance specifc methods that don't get inherited */
}
static void
function_base_class_finalize (HstFunctionClass *class)
{
}


static void
function_class_init (HstFunctionClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (class);

    parent_class = g_type_class_ref (G_TYPE_OBJECT);


    gobject_class->set_property = function_set_property;
    gobject_class->get_property = function_get_property;

    gobject_class->dispose = function_dispose;
    gobject_class->finalize = function_finalize;

    class->eval = function_evaluate;
    class->array=function_array;
    class->resize=function_resize;
    class->range_changed=NULL;

    /**
     * HstFunction:name
     *
     * The name of the function.
     */
    g_object_class_install_property (
        gobject_class,
        PROP_NAME,
        g_param_spec_string ("name", "The name of the function",
                             "The name of the function",
                             "<unnamed>",
                             G_PARAM_WRITABLE | G_PARAM_READABLE )
        );

    /**
     * HstFunction:range
     *
     * The range of the funtion represents any of nx, xmin, xmax. If any
     * of those change, there will be a range-change notification.
     **/
    g_object_class_install_property (
        gobject_class,
        PROP_RANGE,
        g_param_spec_pointer ("range", "The range of the function",
                              "The range of the function.",
                              G_PARAM_WRITABLE | G_PARAM_READABLE )
        );

    /**
     * HstFunction:func
     *
     * A pointer to a double Hst1DFunction function that is to be
     * evaluated by HstFunction
     * 
     */
    g_object_class_install_property (
        gobject_class,
        PROP_FCN,
        g_param_spec_pointer ("func", "function",
                              "Function to evaluate",
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    
    /**
     * HstFunction:nx
     * 
     * The number of bins in the X axis
     */
    g_object_class_install_property (
        gobject_class,
        PROP_NPTS,
        g_param_spec_int("nx","nx","Number of points in X",
                         1,999999,10,
                         G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    
    /**
     * HstFunction:xmin
     * 
     * The minimum value of X definig the range of validity of 
     * the function
     */
    g_object_class_install_property (
        gobject_class,
        PROP_XMIN,
        g_param_spec_double("xmin","xmin","Minimum value of X",
                            -DBL_MAX,DBL_MAX,0.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    
    /**
     * HstFunction:xmax
     * 
     * The maximum value of X definig the range of validity of 
     * the function
     */
    g_object_class_install_property (
        gobject_class,
        PROP_XMAX,
        g_param_spec_double("xmax","xmax","Maximum value of X",
                            -DBL_MAX,DBL_MAX,1.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_XSTEP,
        g_param_spec_double("xstep","xstep","step en X",
                            -DBL_MAX,DBL_MAX,1.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );

    /**
     * HstFunction::range-changed
     *
     * This signal is emitted whenever any of the properties that define
     * the range of the histogram change. That applies to nx, xmin or
     * xmax.
     **/
    function_signals[RANGE_CHANGED]
        = g_signal_new ("range-changed",
                        G_TYPE_FROM_CLASS (gobject_class),
                        G_SIGNAL_RUN_FIRST,
                        G_STRUCT_OFFSET (HstFunctionClass, range_changed),
                        NULL, NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE, 0);


}

static void function_init( HstFunction *f, HstFunctionClass *klass) {
    f->nbin = 1;
    f->xmin=0.;
    f->xmax=1.;
    f->func=0;
    f->name=0;
    g_signal_connect( f, "notify::range",
                      G_CALLBACK(function_notify_range_changed),
                      NULL );
}

static void function_dispose(GObject        *object) {
    g_return_if_fail( HST_IS_FUNCTION(object) );
    G_OBJECT_CLASS (parent_class)->dispose (object);
}
static void function_finalize (GObject        *object) {
    HstFunction *ff;
    g_return_if_fail( HST_IS_FUNCTION(object) );
    ff = HST_FUNCTION(object);
    if (ff)
        g_free( ff->name );
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
function_set_property (GObject      *object,
                       guint         property_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
    HstFunction *ff;
    g_return_if_fail( HST_IS_FUNCTION(object) );

    ff = HST_FUNCTION(object);
    switch (property_id) {
    case PROP_NAME:
        if (ff->name)
            g_free(ff->name);
        ff->name = g_value_dup_string (value);
        break;
    case PROP_RANGE:
        break;
    case PROP_FCN:
        ff->func = (double (*)(double))g_value_get_pointer (value);
        break;
    case PROP_NPTS:
        ff->nbin = g_value_get_int(value);
        ff->step = (ff->xmax-ff->xmin)/((double)ff->nbin);
        g_object_notify(object,"range");
        break;
    case PROP_XMIN:
        ff->xmin = g_value_get_double(value);
/*      if ( ff->xmin>ff->xmax ) */
/*          __swap_double(ff->xmin,ff->xmax); */
        ff->step = (ff->xmax-ff->xmin)/((double)ff->nbin);
        g_object_notify(object,"range");
        break;
    case PROP_XMAX:
        ff->xmax = g_value_get_double(value);
/*      if ( ff->xmin>ff->xmax ) */
/*          __swap_double(ff->xmin,ff->xmax); */
        ff->step = (ff->xmax-ff->xmin)/((double)ff->nbin);
        g_object_notify(object,"range");
        break;
    case PROP_XSTEP:
        ff->step = g_value_get_double(value);
        ff->nbin = (int)( (ff->xmax-ff->xmin)/ff->step );
        ff->xmax = ff->xmin + ff->nbin*ff->step;
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void
function_get_property (GObject     *object,
                       guint        property_id,
                       GValue      *value,
                       GParamSpec  *pspec)
{
    HstFunction *ff;
    g_return_if_fail( HST_IS_FUNCTION(object) );

    ff = HST_FUNCTION(object);

    switch (property_id) {
    case PROP_NAME:
        g_value_set_string (value, ff->name);
        break;
    case PROP_RANGE:
        g_value_set_pointer (value,NULL);
        break;
    case PROP_FCN:
        g_value_set_pointer (value,(void *)ff->func);
        break;
    case PROP_NPTS:
        g_value_set_int(value,ff->nbin);
        break;
    case PROP_XMIN:
        g_value_set_double(value,ff->xmin);
        break;
    case PROP_XMAX:
        g_value_set_double(value,ff->xmax);
        break;
    case PROP_XSTEP:
        g_value_set_double(value,ff->step);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

/**
 * hst_function_new:
 * @n: number of bins
 * @xmin: minimum value of the range
 * @xmax: maximum value of the range
 * @ptr: a pointer to a real value function
 *
 * This is a convenient function to create a #HstFunction and setting
 * its properties in one go.
 *
 * Return value: a #HstFunction object
 **/
HstFunction *hst_function_new(int n, double xmin, double xmax,double (*ptr)(double)) {
    HstFunction *f;
    g_return_val_if_fail ( n>0, NULL );
    g_return_val_if_fail ( xmin!=xmax, NULL );
    f = g_object_new (HST_TYPE_FUNCTION,
                      "nx",n,"xmin",xmin,"xmax",xmax,"func",ptr,NULL);


    return f;
}

static int function_resize(HstFunction *f,int width, int height) {
    g_return_val_if_fail ( HST_FUNCTION(f), 0 );
    return 0;
}

/**
 * hst_function_resize:
 * @f: a #HstFunction object
 * @width: the new width
 * @height: the new height
 *
 * This function is used to change the size of the graphical
 * representation of the function
 *
 * Return value: an integer
 **/
int hst_function_resize(HstFunction *f,int width, int height) {
    return HST_FUNCTION_GET_CLASS(f)->resize(f,width,height);
}

/**
 * hst_function_eval:
 * @f: a #HstFunction object
 * @x: value of x at which we want to evaluate the function
 *
 * Return value: a double. The value of the function at x: f(x)
 **/
double hst_function_eval(HstFunction *f,double x) {
    return HST_FUNCTION_GET_CLASS(f)->eval(f,x);
}

static double function_evaluate(HstFunction *f, double x) {
    if (f->func)
        return f->func(x);
    else
        return 0.;
}

static void function_array(HstFunction *f, Tuple *t, double xmin, double xmax) {
    int i,npts;
    double x;

    t->n = 0;
    t->P = 0;

    g_return_if_fail(HST_IS_FUNCTION(f));

    npts = (xmax-xmin)/f->step;

    t->n = npts;
    if (npts<=0) {
        t->P=0;
        t->n=0;
        return;
    }

    t->P = g_new(PointList,npts);

    for (i=0;i<npts;i++) {
        x = xmin + i*f->step;
        t->P[i].x = x;
        t->P[i].y = hst_function_eval(f,x);
    }
    return;
}

/**
 * hst_function_get_array:
 * @f: a HstFunction
 * @t: a #Tuple
 * @xmin: minimum value of the x range
 * @xmax: maximum value of the x range
 *
 * This function returns an array or #Tuple with the values of @f
 * between @xmin and @xmax
 **/
void hst_function_get_array(HstFunction *f, Tuple *t, double xmin, double xmax) {
    HST_FUNCTION_GET_CLASS(f)->array(f,t,xmin,xmax);
}

/**
 * hst_function_clear_tuple:
 * @t: the #Tuple to be cleared
 *
 * This function is used to release the memory allocated for the array
 * returned by #hst_function_get_array
 *
 */
void hst_function_clear_tuple(Tuple *t) {
    g_free(t->P);
    t->P=0;
    t->n=0;
}

/**
 * hst_function_npts:
 * @f: a #HstFunction
 *
 * Return value: the number of points of the function
 *
 **/
int
hst_function_npts(HstFunction *f)
{
    return f->nbin;
}

/**
 * hst_function_xmin:
 * @f: a #HstFunction
 *
 * Return value: the lower value of the function range
 *
 **/
double
hst_function_xmin(HstFunction *f)
{
    return f->xmin;
}

/**
 * hst_function_xmax:
 * @f: a #HstFunction
 *
 * Return value: the upper value of the function range
 *
 **/
double
hst_function_xmax(HstFunction *f)
{
    return f->xmax;
}

/**
 * hst_function_step:
 * @f: a #HstFunction
 *
 * Return value: the step or binwidth of the function
 *
 **/
double
hst_function_step(HstFunction *f)
{
    return f->step;
}


/**
 * hst_function_find_bin:
 * @f: a #HstFunction
 * @x: the value
 *
 * Return value: the bin that includes @x
 *
 **/
int hst_function_find_bin(HstFunction *f, double x) {
    int wbin;
    g_return_val_if_fail ( HST_FUNCTION(f), -1 );

    wbin = (int)( (x-f->xmin)/f->step );
    if ( wbin<0 )
        wbin = -1;
    if ( wbin>=f->nbin )
        wbin = f->nbin;

    return wbin;
}

/**
 * hst_function_bin_center:
 * @f: a #HstFunction
 * @ib: bin number
 *
 * Return value: the center of the i-th bin
 *
 **/
double     hst_function_bin_center(HstFunction *f, int ib) {
    g_return_val_if_fail ( HST_FUNCTION(f), DBL_MAX );
    return f->xmin + (0.5+ib)*f->step;
}

static void function_notify_range_changed(GObject *o, GParamSpec *arg, gpointer data) {
    g_signal_emit (o,function_signals[RANGE_CHANGED],0);
}

static void function_real_range_changed(HstFunction *f) {
    g_print("function_real_range_changed\n");
}

/**
 * hst_function_get_name:
 * @f: a HstFunction
 *
 * Return value: the name of the function
 **/
gchar     *hst_function_get_name(HstFunction *f) {
    gchar *name;
    g_return_val_if_fail ( HST_FUNCTION(f), NULL );
    g_object_get(G_OBJECT(f),"name",&name,NULL);
    return name;
}


/**
 * hst_function_set_name:
 * @f: a HstFunction
 * @name: the new name for the function
 *
 * Sets a name for a function
 *
 **/
void hst_function_set_name(HstFunction *f, gchar *name) {
    g_return_if_fail ( HST_FUNCTION(f) );
    g_object_set(G_OBJECT(f),"name",name,NULL);
}
