#include <stdio.h>
//#include <values.h>
#include "function2d.h"

enum {
  DESTROY,
  LAST_SIGNAL
};
enum {
  PROP_0,
  PROP_NPTS,
  PROP_NY,
  PROP_YMIN,
  PROP_YMAX,
  PROP_YSTEP,
  PROP_FCN2D
};

#define __swap_double(x,y) {			\
    double tmp = x;				\
    x = y;					\
    y = tmp;					\
}

static void       function2d_base_class_init     (HstFunction2DClass *class);
static void       function2d_base_class_finalize (HstFunction2DClass *class);
static void       function2d_class_init          (HstFunction2DClass *klass);
static void       function2d_init                (HstFunction2D      *object,
					     HstFunction2DClass *klass);

static void	  function2d_set_property	  (GObject	 *object,
					   guint           property_id,
					   const GValue   *value,
					   GParamSpec     *pspec);
static void	  function2d_get_property	  (GObject	 *object,
					   guint           property_id,
					   GValue         *value,
					   GParamSpec     *pspec);
static void       function2d_dispose             (GObject        *object);
static void       function2d_finalize            (GObject        *object);
static double     function2d_evaluate(HstFunction2D *f, double x,double y);

static gpointer    parent_class = NULL;
/* static guint       object_signals[LAST_SIGNAL] = { 0 }; */


GType function2d_get_type (void) {
  static GType object_type = 0;

  if (!object_type) {
      static const GTypeInfo object_info = {
	  sizeof (HstFunction2DClass),
	  (GBaseInitFunc) function2d_base_class_init,
	  (GBaseFinalizeFunc) function2d_base_class_finalize,
	  (GClassInitFunc) function2d_class_init,
	  NULL,		/* class_finalize */
	  NULL,		/* class_data */
	  sizeof (HstFunction2D),
	  16,		/* n_preallocs */
	  (GInstanceInitFunc) function2d_init,
	  NULL,		/* value_table */
      };
      
      object_type = g_type_register_static (HST_TYPE_FUNCTION,
					    "HstFunction2D", 
					    &object_info, 0);
  }

  return object_type;
}

static void
function2d_base_class_init (HstFunction2DClass *class) {
    /* reset instance specifc methods that don't get inherited */
}

static void
function2d_base_class_finalize (HstFunction2DClass *class)
{
}

static void
function2d_class_init (HstFunction2DClass *class)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (class);

  parent_class = g_type_class_ref (HST_TYPE_FUNCTION);


  gobject_class->set_property = function2d_set_property;
  gobject_class->get_property = function2d_get_property;

  gobject_class->dispose = function2d_dispose;
  gobject_class->finalize = function2d_finalize;

  class->eval2d = function2d_evaluate;

#if   GLIB_CHECK_VERSION(2,3,0) 
  g_object_class_override_property (
      gobject_class,
      PROP_FCN2D,
      "func"
      );
#else
  g_object_class_install_property (
      gobject_class,
      PROP_FCN2D,
      g_param_spec_pointer ("func", "function",
			    "Function to evaluate",
			    G_PARAM_READABLE | G_PARAM_WRITABLE)
      );

#endif

  g_object_class_install_property(
      gobject_class,
      PROP_NPTS,
      g_param_spec_int("npts","npts","Number of points (nx x ny)",
		       1,999999,10,
		       G_PARAM_READABLE )
      );
  
  
  g_object_class_install_property (
      gobject_class,
      PROP_NY,
      g_param_spec_int("ny","ny","Number of points in Y",
		       1,999999,10,
		       G_PARAM_READABLE | G_PARAM_WRITABLE)
      );
  g_object_class_install_property (
      gobject_class,
      PROP_YMIN,
      g_param_spec_double("ymin","ymin","Minimum value of Y",
			  -DBL_MAX,DBL_MAX,0.,
			  G_PARAM_READABLE | G_PARAM_WRITABLE)
      );
  g_object_class_install_property (
      gobject_class,
      PROP_YMAX,
      g_param_spec_double("ymax","ymax","Maximum value of Y",
			  -DBL_MAX,DBL_MAX,1.,
			  G_PARAM_READABLE | G_PARAM_WRITABLE)
      );
  g_object_class_install_property (
      gobject_class,
      PROP_YSTEP,
      g_param_spec_double("ystep","ystep","step in Y",
			  -DBL_MAX,DBL_MAX,1.,
			  G_PARAM_READABLE | G_PARAM_WRITABLE)
      );
}

static void function2d_init( HstFunction2D *f, HstFunction2DClass *klass) {
    f->Y = g_object_new( HST_TYPE_FUNCTION,
			 "nx",f->X.nbin,
			 "xmin",f->X.xmin,
			 "xmax",f->X.xmax,
			 "func",NULL,
			 NULL);
}

static void function2d_dispose(GObject        *object) {
    g_return_if_fail( HST_IS_FUNCTION2D(object) );
    G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void function2d_finalize (GObject        *object) {
    g_return_if_fail( HST_IS_FUNCTION2D(object) );
    g_object_unref( HST_FUNCTION2D(object)->Y );
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
function2d_set_property (GObject      *object,
		  guint         property_id,
		  const GValue *value,
		  GParamSpec   *pspec)
{
    HstFunction2D *ff;
    g_return_if_fail( HST_IS_FUNCTION2D(object) );

    ff = HST_FUNCTION2D(object);
    switch (property_id) {
        case PROP_FCN2D:
            ff->func2d = (double (*)(double, double))g_value_get_pointer (value);
            break;
        case PROP_NY:
            g_object_set(G_OBJECT(ff->Y),"nx",g_value_get_int(value),NULL);
            break;
        case PROP_YMIN:
            g_object_set(G_OBJECT(ff->Y),"xmin",g_value_get_double(value),NULL);
            break;
        case PROP_YMAX:
            g_object_set(G_OBJECT(ff->Y),"xmax",g_value_get_double(value),NULL);
            break;
        case PROP_YSTEP:
            g_object_set(G_OBJECT(ff->Y),"xstep",g_value_get_double(value),NULL);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
function2d_get_property (GObject     *object,
		  guint        property_id,
		  GValue      *value,
		  GParamSpec  *pspec)
{
    HstFunction2D *ff;
    g_return_if_fail( HST_IS_FUNCTION2D(object) );

    ff = HST_FUNCTION2D(object);

    switch (property_id) {
        case PROP_FCN2D:
            g_value_set_pointer (value,(void *)ff->func2d);
            break;
        case PROP_NPTS:
            g_value_set_int(value,ff->X.nbin*ff->Y->nbin);
            break;
        case PROP_NY:
            g_value_set_int(value,ff->Y->nbin);
            break;
        case PROP_YMIN:
            g_value_set_double(value,ff->Y->xmin);
            break;
        case PROP_YMAX:
            g_value_set_double(value,ff->Y->xmax);
            break;
        case PROP_YSTEP:
            g_value_set_double(value,ff->Y->step);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}


HstFunction2D *hst_function2d_new(int nx, double xmin, double xmax, int ny, double ymin, double ymax,Hst2DFunction ptr) {    
    HstFunction2D *f;
    g_return_val_if_fail ( nx>0, NULL );
    g_return_val_if_fail ( xmin!=xmax, NULL );
    g_return_val_if_fail ( ny>0, NULL );
    g_return_val_if_fail ( ymin!=ymax, NULL );
    f = g_object_new (HST_TYPE_FUNCTION2D,
                      "nx",nx,"xmin",xmin,"xmax",xmax,
                      "ny",ny,"ymin",ymin,"ymax",ymax,
                      "func",ptr,NULL);
    return f;
}


double hst_function2d_eval(HstFunction2D *f,double x, double y) {
    return HST_FUNCTION2D_GET_CLASS(f)->eval2d(f,x,y);
}

static double function2d_evaluate(HstFunction2D *f, double x, double y) {
    if (f->func2d)
	return f->func2d(x,y);
    else
	return 0.;
}

int hst_function2d_find_bin_from_cell(HstFunction2D *h2d, int ix,int iy) {
    return ix + h2d->X.nbin*iy;
}
void hst_function2d_find_cell_from_bin(HstFunction2D *h2d, int ib , int *ix, int*iy) {
    *ix = ib % h2d->X.nbin;
    *iy = ib / h2d->X.nbin;
}

HstFunction    *hst_function2d_X(HstFunction2D *h2d) {
     return HST_FUNCTION2D_X(h2d);
}
HstFunction    *hst_function2d_Y(HstFunction2D *h2d) {
     return HST_FUNCTION2D_Y(h2d);
}
