#include <stdio.h>
#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include "gtkhistogram.h"
#include "gtkhistogram2d.h"
#include "histogram.h"
#include "function-protected.h"
#include "gtkhistogram-protected.h"
#include "axis.h"

static char line[128];
const static double eps = 0.00001;

static void     hst1d_class_init  (GtkHst1DClass  *klass);
static void     hst1d_init        (GtkHst1D       *hst1d);
static void     hst1d_dispose     (GObject      *object);
static void     hst1d_finalize    (GObject      *object);
static void     hst1d_realize     (GtkWidget *widget);
static gboolean hst1d_expose      (GtkWidget      *widget,
                                   GdkEventExpose *event );
static void hst1d_size_request    (GtkWidget *widget,
                                   GtkRequisition *requisition);
static void hst1d_size_allocate   (GtkWidget     *widget,
                                   GtkAllocation *allocation);

static void hst1d_set_property        (GObject   *object,
                                       guint           property_id,
                                       const GValue   *value,
                                       GParamSpec     *pspec);
static void hst1d_get_property        (GObject   *object,
                                       guint           property_id,
                                       GValue         *value,
                                       GParamSpec     *pspec);
static void hst1d_draw(GtkHst1D *hst);
static void hst1d_draw_frame(GtkHst1D *h);
static void hst1d_draw_stats(GtkHst1D *h);
static void hst1d_draw_pixmap(GtkHst1D *h);
static gboolean hst1d_update(GtkHst1D *hst);
static void hst1d_draw_axis(GtkHst1D *hst);
static gboolean hst1d_button_press(GtkHst1D *, GdkEventButton *, gpointer *);
static gboolean hst1d_button_release(GtkHst1D *, GdkEventButton *, gpointer *);
static gboolean hst1d_motion_notify(GtkHst1D *, GdkEventMotion *, gpointer *);
static gboolean hst1d_key_press(GtkHst1D *, GdkEventKey *, gpointer *);
static GtkWidget *hst1d_create_menu( GtkHst1D *h);
static void     hst1d_set_style_colors( GtkHst1D *h);
static void     hst1d_set_curve_colors( GtkHst1D *h);
static void hst1d_set_color(GtkHst1D *h, int i, GdkColor *col);
static void hst1d_range_changed(HstFunction *f,GtkHst1D *h);
static void hst1d_unrealize(GtkWidget *);
static gpointer parent_class;

#define HST_MARGIN_W(h) (4*h->cnvs->cw)
#define HST_MARGIN_H(h) (4*h->cnvs->ch)
#define LEGEND_XTRA 0.25
enum {
    PROP_0,
    PROP_FUNC,
    PROP_XTITLE,
    PROP_YTITLE,
    PROP_LOGX,
    PROP_LOGY,
    PROP_FILLED,
    PROP_STATS,
    PROP_HXMAX,
    PROP_HXMIN,
    PROP_HYMAX,
    PROP_HYMIN,
    PROP_VMAX,
    PROP_VMIN,
    PROP_CCOLORS,
    PROP_PIXMAP,
};

static void  draw_axis (GtkHst1D *,GtkHstAxisType, int, int logx);

typedef enum
{
    HST_BUTTON_PRESS=1,
    HST_BUTTON_RELEASE=2,
    HST_KEY_PRESS=4
} GtkHst1DButtonEventType;

struct _GtkHst1DEvent
{
    guint key;
    guint button;
    double x,y;
    double x0,y0;
};

GType gtk_hst_colors_get_type(void)
{
    static GType etype = 0;
    if (etype == 0)
    {
        static const GEnumValue values[] = {
            { HST_COLOR_FOREGROUND,"HST_COLOR_FOREGROUND","foreground-color"},
            { HST_COLOR_BACKGROUND,"HST_COLOR_BACKGROUND","background-color"},
            { HST_COLOR_AXIS,"HST_COLOR_AXIS","axis-color"},
            { 0, NULL, NULL }
        };
        etype = g_enum_register_static ("GtkHstColors", values);
    }
    return etype;
}


GType gtk_hst_axis_pos_get_type (void)
{
    static GType etype = 0;
    if (etype == 0) {
        static const GEnumValue values[] = {
            {HST_AXIS_NONE,"HST_AXIS_NONE","none"},
            {HST_AXIS_LEFT,"HST_AXIS_LEFT","left"},
            {HST_AXIS_MIDDLE,"HST_AXIS_MIDDLE","middle"},
            {HST_AXIS_RIGHT,"HST_AXIS_RIGHT","right"},
            {HST_AXIS_X_MASK,"HST_AXIS_X_MASK","x-mask"},
            {HST_AXIS_Y_MASK,"HST_AXIS_Y_MASK","y-mask"},
            {HST_AXIS_Y_MASK,"HST_AXIS_Z_MASK","z-mask"},
            { 0, NULL, NULL }
        };
        etype = g_enum_register_static ("GtkHstAxisPos", values);

    }
    return etype;
}

GType gtk_hst_attr_get_type(void)
{
    static GType etype = 0;
    if (etype == 0)
    {
        static const GEnumValue values[] = {
            { HST_FILLED,"HST_FILLED","filled"},
            { HST_LOGX,"HST_LOGX","logx"},
            { HST_LOGY,"HST_LOGY","logy"},
            { HST_STATS, "HST_STATS","stats"},
            { 0, NULL, NULL }
        };
        etype = g_enum_register_static ("GtkHstAttr", values);
    }
    return etype;
}

GType gtk_hst_axis_type(void)
{
    static GType etype = 0;
    if (etype == 0)
    {
        static const GEnumValue values[] = {
            { HST_X_AXIS,"HST_X_AXIS","x-axis"},
            { HST_Y_AXIS,"HST_Y_AXIS","y-axis"},
            { HST_Z_AXIS,"HST_Z_AXIS","z-axis"},
            { 0, NULL, NULL }
        };
        etype = g_enum_register_static ("GtkHstAxisType", values);
    }
    return etype;

}

GType gtk_hst1d_get_type (void)
{
    static GType hst1d_type = 0;

    if (!hst1d_type)
    {
        static const GTypeInfo hst1d_info = {
            sizeof (GtkHst1DClass),
            NULL,               /* base_init */
            NULL,               /* base_finalize */
            (GClassInitFunc) hst1d_class_init,
            NULL,               /* class_finalize */
            NULL,               /* class_data */
            sizeof (GtkHst1D),
            0,          /* n_preallocs */
            (GInstanceInitFunc) hst1d_init,
        };

        hst1d_type = g_type_register_static (GTK_TYPE_WIDGET, "GtkHst1D",
                                             &hst1d_info, 0);
    }

    return hst1d_type;
}

static void
hst1d_class_init (GtkHst1DClass *class)
{
    GObjectClass *gobject_class;
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;

    parent_class = g_type_class_peek_parent (class);

    gobject_class = G_OBJECT_CLASS (class);
    object_class = GTK_OBJECT_CLASS (class);

    gobject_class->dispose = hst1d_dispose;
    gobject_class->finalize = hst1d_finalize;
    gobject_class->set_property = hst1d_set_property;
    gobject_class->get_property = hst1d_get_property;

    widget_class = GTK_WIDGET_CLASS (class);

    widget_class->realize = hst1d_realize;
    widget_class->unrealize = hst1d_unrealize;
    widget_class->expose_event = hst1d_expose;
    widget_class->size_request = hst1d_size_request;
    widget_class->size_allocate = hst1d_size_allocate;

    class->draw_pixmap=hst1d_draw_pixmap;
    class->draw_frame =hst1d_draw_frame;
    class->draw_stats=hst1d_draw_stats;
    class->draw_axis=hst1d_draw_axis;
    class->xtra_menu = NULL;
    
    /**
     * GtkHst1D:func
     * 
     * The pointer of the main function to be displayed in the widget. 
     * This is the function that will fix the range of the axis. 
     * 
     */
    g_object_class_install_property (
        gobject_class,
        PROP_FUNC,
        g_param_spec_pointer ("func", "function",
                              "Function to evaluate",
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );


    g_object_class_install_property (
        gobject_class,
        PROP_XTITLE,
        g_param_spec_string ("xtitle", "X axis label",
                             "Label for X axis",
                             "X",
                             G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_YTITLE,
        g_param_spec_string ("ytitle", "Y axis label",
                             "Label for Y axis",
                             "Y",
                             G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_LOGX,
        g_param_spec_boolean ("logx", "X axis in log scale",
                              "If set shows X axis in log scale",
                              FALSE,
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_LOGY,
        g_param_spec_boolean ("logy", "Y axis in log scale",
                              "If set shows Y axis in log scale",
                              FALSE,
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_FILLED,
        g_param_spec_boolean ("filled", "Filled hst1ds",
                              "If set draws a filled histogram",
                              FALSE,
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_STATS,
        g_param_spec_boolean ("stats", "Show stats",
                              "If set shows the statistics",
                              FALSE,
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );

    g_object_class_install_property (
        gobject_class,
        PROP_HXMIN,
        g_param_spec_double("hxmin","hxmin","Minimum value of X in the histogram range",
                            -DBL_MAX,DBL_MAX,0.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_HXMAX,
        g_param_spec_double("hxmax","hxmax","Maximum value of X in the histogram range",
                            -DBL_MAX,DBL_MAX,1.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_HYMIN,
        g_param_spec_double("hymin","hymin","Minimum value of Y in the histogram range",
                            -DBL_MAX,DBL_MAX,0.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_HYMAX,
        g_param_spec_double("hymax","hymax","Maximum value of Y in the histogram range",
                            -DBL_MAX,DBL_MAX,1.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );


    g_object_class_install_property (
        gobject_class,
        PROP_VMIN,
        g_param_spec_double("vmin","vmin","Minimum value to be shown",
                            -DBL_MAX,DBL_MAX,0.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    g_object_class_install_property (
        gobject_class,
        PROP_VMAX,
        g_param_spec_double("vmax","vmax","Maximum value to be shown",
                            -DBL_MAX,DBL_MAX,1.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE)
        );

    g_object_class_install_property(
        gobject_class,
        PROP_PIXMAP,
        g_param_spec_object("pixmap","pixmap",
                            "Pixbuffer",
                            GDK_TYPE_PIXBUF,
                            G_PARAM_READABLE
            )
        );
}

static void
hst1d_set_property (GObject      *object,
                    guint         property_id,
                    const GValue *value,
                    GParamSpec   *pspec)
{
    GtkHst1D *h;
    HstFunction *f;
    g_return_if_fail( GTK_IS_HST1D(object) );


    h = GTK_HST1D(object);
    switch (property_id)
    {
    case PROP_FUNC:
        f = HST_FUNCTION(g_value_get_pointer (value));
        if ( f && f!=h->func )
        {
            if (h->func)
            {
                h->curves = g_list_remove( h->curves, h->func );
                if (h->sync_id>=0)
                    g_signal_handler_disconnect(h->func,h->sync_id);
                g_object_unref(h->func);
            }
            h->func = f;
            if ( h->sync_id>=0 )
                h->sync_id = g_signal_connect(h->func,"range-changed",
                                              G_CALLBACK(hst1d_range_changed),
                                              h
                    );
            g_object_ref(h->func);
            h->curves = g_list_prepend(h->curves, h->func);
            h->xmin = f->xmin;
            h->xmax = f->xmax;
            gtk_widget_queue_resize( GTK_WIDGET(object) );
        }
        break;
    case PROP_XTITLE:
        if ( h->xtitle )
            g_free( h->xtitle);
        h->xtitle = g_strdup( g_value_get_string (value) );
        break;
    case PROP_YTITLE:
        if ( h->ytitle )
            g_free( h->ytitle);
        h->ytitle = g_strdup( g_value_get_string (value) );
        break;
    case PROP_LOGX:
        gtk_hst1d_logx(h,g_value_get_boolean(value));
        break;
    case PROP_LOGY:
        gtk_hst1d_logy(h,g_value_get_boolean(value));
        break;
    case PROP_FILLED:
        gtk_hst1d_filled(h,g_value_get_boolean(value));
        break;
    case PROP_STATS:
        gtk_hst1d_show_stats(h,g_value_get_boolean(value));
        break;
    case PROP_VMIN:
    case PROP_HYMIN:
        h->vmin = h->cnvs->ymin = g_value_get_double(value);
        if (h->cnvs->ymin==h->cnvs->ymax)
            h->cnvs->fixed=FALSE;
        else
            h->cnvs->fixed=TRUE;
        break;
    case PROP_VMAX:
    case PROP_HYMAX:
        h->vmax = h->cnvs->ymax = g_value_get_double(value);
        if (h->cnvs->ymin==h->cnvs->ymax)
            h->cnvs->fixed=FALSE;
        else
            h->cnvs->fixed=TRUE;
        break;
    case PROP_HXMIN:
        h->xmin = h->cnvs->xmin = g_value_get_double(value);
        gtk_widget_queue_resize( GTK_WIDGET(h) );
        break;
    case PROP_HXMAX:
        h->xmax = h->cnvs->xmax = g_value_get_double(value);
        gtk_widget_queue_resize( GTK_WIDGET(h) );
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void
hst1d_get_property (GObject      *object,
                    guint         property_id,
                    GValue       *value,
                    GParamSpec   *pspec)
{
    GtkHst1D *h;
    g_return_if_fail( GTK_IS_HST1D(object) );

    h = GTK_HST1D(object);
    switch (property_id)
    {
    case PROP_FUNC:
        g_value_set_pointer (value,(void *)h->func);
        break;
    case PROP_XTITLE:
        g_value_set_string( value, h->xtitle );
        break;
    case PROP_YTITLE:
        g_value_set_string( value, h->ytitle );
        break;
    case PROP_LOGX:
        g_value_set_boolean( value, gtk_hst1d_is_logx(h) );
        break;
    case PROP_LOGY:
        g_value_set_boolean( value, gtk_hst1d_is_logy(h) );
        break;
    case PROP_FILLED:
        g_value_set_boolean( value, gtk_hst1d_is_filled(h) );
        break;
    case PROP_STATS:
        g_value_set_boolean( value, gtk_hst1d_stats(h));
        break;
    case PROP_HYMIN:
        g_value_set_double(value,h->cnvs->ymin);
        break;
    case PROP_HYMAX:
        g_value_set_double(value,h->cnvs->ymax);
        break;
    case PROP_HXMIN:
        g_value_set_double(value,h->cnvs->xmin);
        break;
    case PROP_HXMAX:
        g_value_set_double(value,h->cnvs->xmax);
        break;
    case PROP_VMIN:
        g_value_set_double(value,h->vmin);
        break;
    case PROP_VMAX:
        g_value_set_double(value,h->vmax);
        break;
    case PROP_PIXMAP:
    {
        GdkPixbuf *pxb = gdk_pixbuf_get_from_drawable(
            NULL,
            h->pm,
            NULL,
            0,0,0,0,-1,-1);
        g_value_set_object_take_ownership(value,pxb);
        break;
    }

    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}


static void
hst1d_init (GtkHst1D *h)
{
    h->cnvs = g_new0(Canvas,1);
    h->evt  = g_new(GtkHst1DEvent,1);
    memset(h->evt,0,sizeof(GtkHst1DEvent));
    memset(h->cnvs,0,sizeof(Canvas));
    h->cnvs->stats = hst_stats_new();
    h->gc = 0;
    h->pm = 0;
    h->xtitle = 0;
    h->ytitle = 0;
    h->vmin=h->vmax=0.;
    h->drawaxis = HST_AXIS_LEFT << 4 | HST_AXIS_LEFT;
    h->attr = 0;
    h->colors_changed = FALSE;
    h->tmout = h->sync_id = -1;
    gtk_hst1d_set_colors(h,"foreground","black",
                         "background","#E6E6E6",
                         "axis","black",NULL);

    gtk_widget_set_events (GTK_WIDGET (h),
                           GDK_BUTTON_PRESS_MASK |
                           GDK_BUTTON_RELEASE_MASK |
                           GDK_KEY_PRESS_MASK |
                           GDK_POINTER_MOTION_MASK );

    gtk_signal_connect (GTK_OBJECT (h), "button_press_event",
                        GTK_SIGNAL_FUNC (hst1d_button_press),
                        NULL);
    gtk_signal_connect (GTK_OBJECT (h), "button_release_event",
                        GTK_SIGNAL_FUNC (hst1d_button_release),
                        NULL);
    gtk_signal_connect (GTK_OBJECT (h), "motion_notify_event",
                        GTK_SIGNAL_FUNC (hst1d_motion_notify),
                        NULL);
    gtk_signal_connect (GTK_OBJECT (h), "key_press_event",
                        GTK_SIGNAL_FUNC (hst1d_key_press),
                        NULL);


}

static void
hst1d_finalize (GObject *object)
{
    GList *tmp_list;
    GtkHst1D *hst1d = GTK_HST1D (object);

    if (hst1d->tmout>=0)
        gtk_hst1d_monitor_stop (hst1d);

    g_object_unref( hst1d->cnvs->stats);
    g_free(hst1d->cnvs);
    if (hst1d->xtitle)
    {
        g_free (hst1d->xtitle);
        hst1d->xtitle = NULL;
    }
    if (hst1d->ytitle)
    {
        g_free (hst1d->ytitle);
        hst1d->ytitle = NULL;
    }

    for(tmp_list=g_list_first(hst1d->curves);tmp_list;)
    {
        g_object_unref( G_OBJECT(tmp_list->data) );
        tmp_list = g_list_next (tmp_list);
    }
    if ( hst1d->pm )
    {
        g_object_unref (hst1d->pm);
        hst1d->pm = NULL;
    }

    if ( hst1d->cnvs->legend )
        g_object_unref( hst1d->cnvs->legend );

    G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
hst1d_dispose (GObject *object)
{
    G_OBJECT_CLASS (parent_class)->dispose (object);
}

/**
 * gtk_hst1d_new:
 * @ff : A #HstFunction object to be drawn.
 *       This is the object that will define the range of the axis. 
 * Return value: a #GtkWidget
 * 
 */
GtkWidget*
gtk_hst1d_new (HstFunction *ff)
{
    GtkWidget *w;

    g_return_val_if_fail ( ff!=0, NULL);


    w = g_object_new (GTK_TYPE_HST1D,
                      "func",ff,
                      "hxmin",ff->xmin,
                      "hxmax",ff->xmax,
                      NULL);
    return w;
}

/**
 * gtk_hst1d_get_func:
 * @h : A #GtkHst1D object
 * @n : the index of the function to be retrieved
 * 
 * Return value: The #HstFunction object that is stored at 
 * position @n in the list of function to be drawn. 
 */
HstFunction *
gtk_hst1d_get_func(GtkHst1D *h, int n)
{
    g_return_val_if_fail ( h!=NULL, NULL);
    g_return_val_if_fail ( GTK_IS_HST1D(h), NULL);

    return g_list_nth_data (h->curves,n);
}

/**
 * gtk_hst1d_get_n_func:
 * @h: A #GtkHst1D object
 * 
 * Return value: the number of functions to be drawn.
 */
int
gtk_hst1d_get_n_func(GtkHst1D *h)
{
    g_return_val_if_fail ( h!=NULL, 0);
    g_return_val_if_fail ( GTK_IS_HST1D(h), 0);

    return g_list_length(h->curves);
}

/**
 * gtk_hst1d_add_curve:
 * @h: A #GtkHst1D object
 * @f: A #HstFunction object that will added to the list of
 *     functions to be drawn. 
 * 
 * It adds a new function to the list. However, only the points in the
 * range of the main function will be show.
 */
void
gtk_hst1d_add_curve(GtkHst1D *h, HstFunction *f)
{
    g_return_if_fail (h != NULL);
    g_return_if_fail (GTK_IS_HST1D (h));

    if (h->func==0)
    {
        gtk_hst1d_set_func(h,f);
    }
    else
    {
        h->curves = g_list_append(h->curves, f);
        gtk_widget_queue_resize( GTK_WIDGET(h)  );
    }

    g_object_ref(f);
}

/**
 * gtk_hst1d_remove_curves:
 * @h: A #GtkHst1D object
 * 
 * Clears the list of functions to draw
 */
void
gtk_hst1d_remove_curves(GtkHst1D *h)
{
    GObject *obj;
    for(h->curves=g_list_first(h->curves);h->curves;)
    {
        obj = h->curves->data;
        h->curves = g_list_remove( h->curves, obj );
        g_object_unref( obj );
    }
    if (h->func)
        g_object_unref(h->func);
    h->func=0;
}

/**
 * gtk_hst1d_set_func:
 * @h: A #GtkHst1D object
 * @f: A #HstFunction object
 * 
 * Sets @f as the main function to draw. THis will be the function
 * defining the range of the axis.
 * 
 */
void
gtk_hst1d_set_func(GtkHst1D *h, HstFunction *f)
{
    g_object_set(G_OBJECT(h),"func",f,NULL);
}

/**
 * gtk_hst1d_draw_x_axis:
 * @h: A #GtkHst1D object
 * @val: 
 * 
 * Specifies where to draw the X axis
 */
void
gtk_hst1d_draw_x_axis (GtkHst1D * hst, GtkHstAxisPos val)
{
    hst->drawaxis = (hst->drawaxis & ~HST_AXIS_X_MASK) | val;
}

/**
 * gtk_hst1d_draw_y_axis:
 * @h: A #GtkHst1D object
 * @val: 
 * 
 * Specifies where to draw the Y axis
 */
void
gtk_hst1d_draw_y_axis (GtkHst1D * hst, GtkHstAxisPos val)
{
    hst->drawaxis = (hst->drawaxis & ~HST_AXIS_Y_MASK) | val << 4;
}

/**
 * gtk_hst1d_filled;
 * @h: A #GtkHst1D object
 * @z: boolean flag
 * 
 * If @f is set, theh function will be drawn as a filled curve,
 * otherwise only a line will be drawn.
 */
void
gtk_hst1d_filled (GtkHst1D *h, gboolean z)
{
    if ( z )
        h->attr |= HST_FILLED;
    else
        h->attr &= ~HST_FILLED;
}

/**
 * gtk_hst1d_is_filled:
 * @h: A #GtkHst1D object
 * 
 * Return value: the value of the filled flag
 * 
 */
gboolean
gtk_hst1d_is_filled (GtkHst1D *h)
{
    return (h->attr & HST_FILLED);
}

/**
 * gtk_hst1d_logx;
 * @h: A #GtkHst1D object
 * @z: boolean flag
 * 
 * If @f is set, theh function will be drawn with logarithmic
 * scale
 */
 void
gtk_hst1d_logx (GtkHst1D *h, gboolean z)
{
    if ( z )
        h->attr |= HST_LOGX;
    else
        h->attr &= ~HST_LOGX;
}

 /**
  * gtk_hst1d_is_logx:
  * @h: A #GtkHst1D object
  * 
  * Return value: if true, logarithmic scale is being used 
  * for the X axis
  * 
  */
gboolean
gtk_hst1d_is_logx (GtkHst1D *h)
{
    return (h->attr & HST_LOGX);
}

/**
 * gtk_hst1d_logy;
 * @h: A #GtkHst1D object
 * @z: boolean flag
 * 
 * If @f is set, the Y axis will use logarithmic scale
 */
void
gtk_hst1d_logy (GtkHst1D *h, gboolean z)
{
    if ( z )
        h->attr |= HST_LOGY;
    else
        h->attr &= ~HST_LOGY;
}

/**
  * gtk_hst1d_is_logy:
  * @h: A #GtkHst1D object
  * 
  * Return value: if true, logarithmic scale is being used 
  * for the Y axis
  * 
  */
gboolean
gtk_hst1d_is_logy (GtkHst1D *h)
{
    return (h->attr & HST_LOGY);
}

/**
 * gtk_hst1d_show_stats;
 * @h: A #GtkHst1D object
 * @z: boolean flag
 * 
 * If @f is set we draw a summary if the statistics, like
 * numbe of entries, mean value, etc.
 */
void
gtk_hst1d_show_stats (GtkHst1D *h, gboolean z)
{
    if ( z )
        h->attr |= HST_STATS;
    else
        h->attr &= ~HST_STATS;
}

/**
 * gtk_hst1d_stats:
 * @h: A #GtkHst1D object
 * 
 * Return value: if true, we are drawing the stats.
 * 
 */
gboolean
gtk_hst1d_stats (GtkHst1D *h)
{
    return (h->attr & HST_STATS);
}

static
void hst1d_realize(GtkWidget *widget)
{
    GtkHst1D *h;
    GdkWindowAttr attributes;
    gint attributes_mask;
    int icol;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_HST1D (widget));
    h = GTK_HST1D(widget);

    GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual (widget);
    attributes.colormap = gtk_widget_get_colormap (widget);
    attributes.event_mask = gtk_widget_get_events (widget) |
        GDK_BUTTON_PRESS_MASK |
        GDK_BUTTON_RELEASE_MASK |
        GDK_KEY_PRESS_MASK |
        GDK_POINTER_MOTION_MASK |
        GDK_EXPOSURE_MASK;

    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
                                     &attributes,
                                     attributes_mask);
    gdk_window_set_user_data (widget->window, h);
    widget->style = gtk_style_attach (widget->style, widget->window);
    gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

    h->pm = gdk_pixmap_new (widget->window,
                            widget->allocation.width,
                            widget->allocation.height,
                            -1);

    for (icol=0;icol<3;icol++)
        hst1d_set_color(h,icol,h->colors+icol);
    hst1d_draw(h);
}

static void
hst1d_unrealize( GtkWidget *w)
{
    g_object_unref( GTK_HST1D(w)->pm );
    GTK_HST1D(w)->pm=0;
    GTK_WIDGET_CLASS (parent_class)->unrealize(w);
}

static gboolean
hst1d_expose(GtkWidget *widget,GdkEventExpose *event )
{
    hst1d_draw(GTK_HST1D(widget));
    return FALSE;
}

static void
hst1d_size_request (GtkWidget      *widget,
                    GtkRequisition *requisition)
{
    GtkHst1D *hst;
    int      width,nbin;
    float    lmin,lmax,binw;
    int      xaxis;
    int      ncurves;
    PangoLayout    *layout;
    PangoRectangle  fsize;
    gchar           cline[16];

    hst = GTK_HST1D (widget);
    xaxis = hst->drawaxis & HST_AXIS_X_MASK;
    layout = gtk_widget_create_pango_layout (widget,"888");
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    if ( xaxis && HST_AXIS_NONE )
        width = 50;
    else
    {
        getbins(hst->xmin,hst->xmax,10,&lmin,&lmax,&nbin,&binw);
        width = (nbin+1)*fsize.width;
    }
    g_object_unref(layout);

    requisition->width  = HST_MARGIN_W(hst) + width;
    requisition->height = HST_MARGIN_H(hst) + (int)(7.5*fsize.height);

    /*
      Check if we need to draw a legend on the right of the
      histogram. That should only happen when we have more than one
      function in the hitogram.

      We start by deleting the existing one, looping on the functions
      and creating a new one if need be.
    */
    if ( hst->cnvs->legend )
    {
        g_object_unref( hst->cnvs->legend );
        hst->cnvs->legend = NULL;
    }
    ncurves = g_list_length( hst->curves );
    if (ncurves>1)
    {
        GList *lst = hst->curves;
        GList *lc = hst->cnvs->curve_colors;
        GdkColor *color;
        PangoRectangle lsize;
        GString  *str = g_string_new("");
        hst->cnvs->legend = gtk_widget_create_pango_layout (widget,NULL);

        if (!lc)
            color = hst->colors+HST_COLOR_FOREGROUND;

        for ( ;lst;lst=lst->next )
        {
            gchar *name;
            HstFunction *ff = HST_FUNCTION(lst->data);
            if (lc)
            {
                color = (GdkColor *)lc->data;
                if (!lc->next)
                    lc = g_list_first(lc);
                else
                    lc = lc->next;
            }
            sprintf(cline,"#%02x%02x%02x",
                    color->red/256,
                    color->green/256,
                    color->blue/256);
            g_object_get(G_OBJECT(ff),"name",&name,NULL);
            if (name==NULL)
                name = g_strdup("[unnamed]");
            g_string_append_printf(
                str,
                "<span foreground=\"%s\">%s</span>\n",
                cline,name);
            g_free(name);
        }
        layout = hst->cnvs->legend = gtk_widget_create_pango_layout (widget,NULL);
        pango_layout_set_markup(layout,str->str,-1);
        pango_layout_get_pixel_extents (layout, NULL,&lsize);
        requisition->width += (int)( (1.+LEGEND_XTRA)*lsize.width);
        if ( requisition->height<lsize.height)
            requisition->height = lsize.height;

        g_string_free(str,TRUE);
    }

    /* Chain up to default that simply reads current requisition */
    GTK_WIDGET_CLASS (parent_class)->size_request (widget, requisition);
}


static void
hst1d_size_allocate(GtkWidget     *widget, GtkAllocation *allocation)
{
    GList *lst;
    GtkHst1D *h= GTK_HST1D (widget);
    int width,height;
    GtkAllocation current;

    /* Get old values */
    current = *allocation;
    width = widget->allocation.width;
    height= widget->allocation.height;

    /* Chain up to default that simply reads current requisition */
    GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);

    width = widget->allocation.width;
    height= widget->allocation.height;
    
    lst = h->curves;
    for ( ;lst;lst=lst->next )
    {
        HstFunction *ff = HST_FUNCTION(lst->data);
        hst_function_resize(ff,width-HST_MARGIN_W(h),height-HST_MARGIN_H(h));
    }
    if ( h->func )
    {
        h->xmin = h->func->xmin;
        h->xmax = h->func->xmax;
    }

    if ( widget->window )
    {
        if (h->pm)
            g_object_unref (h->pm);

        h->pm = gdk_pixmap_new (widget->window,width,height,-1);

//        hst1d_draw(h);
        gtk_widget_queue_draw(GTK_WIDGET(h));
    }
}

/**
 * gtk_hst1d_monitor_start:
 * @h: A #GtkHst1D object
 * @tim: the time interval
 * 
 * It sets up a timeout callback that will re-draw the 
 * functions in the list every @tim milliseconds.
 */
void
gtk_hst1d_monitor_start (GtkHst1D * hst, int tim)
{
    if (hst->tmout>=0)
        gtk_hst1d_monitor_stop (hst);

    hst->tmout = g_timeout_add (tim,
                                (GtkFunction)hst1d_update,
                                hst);
}

/**
 * gtk_hst1d_monitor_stop:
 * @h: A #GtkHst1D object
 * 
 * Stops the automatic re-drawing of the curves
 */
void
gtk_hst1d_monitor_stop (GtkHst1D * hst)
{
    if (hst->tmout >= 0)
        g_source_remove (hst->tmout);
    hst->tmout = -1;
}

static gboolean
hst1d_update(GtkHst1D *hst)
{
    hst1d_draw(hst);
    gtk_widget_queue_draw(GTK_WIDGET(hst));
    return TRUE;
}

/**
 * gtk_hst1d_get_xtitle:
 * @hst: A #GtkHst1D object
 * 
 * Return value: the title or label of the X axis
 */
gchar *
gtk_hst1d_get_xtitle (GtkHst1D * hst)
{
    return hst->xtitle;
}

/**
 * gtk_hst1d_set_xtitle:
 * @hst: A #GtkHst1D object
 * @tit: the new tittle
 * 
 * Sets the label of the X axis
 * 
 */
void
gtk_hst1d_set_xtitle (GtkHst1D * hst, const char *tit)
{
    if (tit == NULL)
        return;
    if (hst->xtitle)
        g_free (hst->xtitle);
    hst->xtitle = g_strdup (tit);
}

/**
 * gtk_hst1d_get_ytitle:
 * @hst: A #GtkHst1D object
 * 
 * Return value: the title or label of the Y axis
 */
gchar *
gtk_hst1d_get_ytitle (GtkHst1D * hst)
{
    return hst->ytitle;
}

/**
 * gtk_hst1d_set_ytitle:
 * @hst: A #GtkHst1D object
 * @tit: the new tittle
 * 
 * Sets the label of the Y axis
 * 
 */
void
gtk_hst1d_set_ytitle (GtkHst1D * hst, const char *tit)
{
    if (tit == NULL)
        return;
    if (hst->ytitle)
        g_free (hst->ytitle);
    hst->ytitle = g_strdup (tit);
}


/**
 * gtk_hst1d_sync_range:
 * @h: A #GtkHst1D object
 * @z: a boolean flag
 * 
 * If @z is true, the widget will recompute the range of the
 * axis and will redraw the curves whenever the the range of 
 * the main #HstFunction changes.
 * 
 */
void
gtk_hst1d_sync_range(GtkHst1D *h, gboolean z)
{
    g_return_if_fail( GTK_IS_HST1D(h) );
    if (z)
    {
        if ( h->sync_id < 0 )
            h->sync_id = g_signal_connect(h->func,"range-changed",
                                          G_CALLBACK(hst1d_range_changed),
                                          h
                );
    }
    else
    {
        if (h->sync_id>=0)
        {
            g_signal_handler_disconnect(h->func,h->sync_id);
            h->sync_id=-1;
        }
    }
}

/**
 * gtk_hst1d_set_curve_colors:
 * @h: A #GtkHst1D object
 * 
 * 
 * 
 */
void
gtk_hst1d_set_curve_colors(GtkHst1D *h,...)
{
    char *A;
    GList *lst;
    GdkColor *color;
    va_list var_args;

    g_return_if_fail( GTK_IS_HST1D(h) );

    va_start(var_args,h);

    // first clear the existing list
    lst = h->cnvs->curve_colors;
    for (;lst;lst=lst->next)
        g_free( lst->data );

    g_list_free(lst);

    // Now fill the new one
    lst = NULL;
    while (1)
    {
        A=va_arg(var_args,char *);
        if (!A)
            break;

        gtk_hst1d_add_curve_color(h,A);
    }
    va_end(var_args);
}

/**
 * gtk_hst1d_add_curve_color:
 * @h: a #GtkHst1D object
 * @color_name: an existing color name
 * 
 * Appends a color with this name to the list of 
 * colors for the variuos functions to draw. The color 
 * name should be recognized by #gdk_color_parse
 * 
 */
void
gtk_hst1d_add_curve_color(GtkHst1D *h, gchar *color_name)
{
    g_return_if_fail( GTK_IS_HST1D(h) );
    GdkColor *color;

    color = g_new(GdkColor,1);
    if ( !gdk_color_parse(color_name,color) )
    {
        g_printerr("add_curve_color: Cannot get color %s\n",color_name);
        g_free(color);
        return;
    }
    if ( GTK_WIDGET_REALIZED(GTK_WIDGET(h)) )
    {
        if ( ! gdk_colormap_alloc_color (
                 gtk_widget_get_colormap (GTK_WIDGET(h)),
                 color,FALSE,FALSE)
            )
        {
            g_printerr("add_curve_color: could not allocate color %s\n",color_name);
            g_free(color);
        }
        else
            h->cnvs->curve_colors = g_list_append( h->cnvs->curve_colors , color);
    }
    else
    {
        h->cnvs->curve_colors = g_list_append( h->cnvs->curve_colors , color);
        h->colors_changed = TRUE;
    }
}


static void
hst1d_set_style_colors( GtkHst1D *h)
{
    int i;
    GtkWidget *wdgt = GTK_WIDGET(h);
    GtkStyle *style = gtk_style_copy (gtk_widget_get_style (wdgt));
    for (i=0;i<5;i++)
    {
        style->fg[i]=h->colors[0];
        style->bg[i]=h->colors[1];
    }
    gtk_widget_set_style(wdgt,style);
}

static void
hst1d_set_curve_colors( GtkHst1D *h)
{
    GList *lst = h->cnvs->curve_colors;
    for (;lst;lst=lst->next)
    {
        GdkColor *color = (GdkColor *)lst->data;
        gdk_colormap_alloc_color (
            gtk_widget_get_colormap (GTK_WIDGET(h)),
            color,FALSE,FALSE);
    }
}

static void
hst1d_set_color(GtkHst1D *h, int i, GdkColor *col)
{
    if ( col != (h->colors+i) )
        h->colors[i] = *col;

    if ( GTK_WIDGET_REALIZED(GTK_WIDGET(h)) )
    {
        if ( ! gdk_colormap_alloc_color (
                 gtk_widget_get_colormap (GTK_WIDGET(h)),
                 h->colors+i,FALSE,FALSE)
            )
        {
            g_printerr("hst1d_set_color: could not allocate color %d\n",i);
        }
    }
    h->colors_changed = TRUE;
}

#define IS_THIS_COLOR(str,color)  (!strcmp(str,color))
/**
 * gtk_hst1d_set_colors:
 * @h: a #GtkHst1D object
 * @...: NULL terminated list of color names.
 * 
 * Sets the colors for the axis, background an foreground of the
 * canvas. The colors are given with names that #gdk_parse_color
 * understands preceded by the item that will receive the color.
 * 
 */
void
gtk_hst1d_set_colors(GtkHst1D *h,...)
{
    gboolean rc;
    gint     wc;
    char    *A,arg[256],color_name[256];
    va_list var_args;
    va_start(var_args,h);

    color_name[0]='\0';
    do
    {
        A = va_arg(var_args,char *);
        if (!A)
            continue;

        strcpy(arg,A);

        A=va_arg(var_args,char *);
        if (!A)
        {
            g_warning("gtk_hst1d_set_colors: no color name given for %s",arg);
            continue;
        }
        strcpy(color_name,A);

        rc=FALSE;
        wc=HST_NCOLORS;
        if ( IS_THIS_COLOR(arg,"foreground") )
        {
            wc = HST_COLOR_FOREGROUND;
        }
        else if ( IS_THIS_COLOR(arg,"background") )
        {
            wc = HST_COLOR_BACKGROUND;
        }
        else if ( IS_THIS_COLOR(arg,"axis") )
        {
            wc = HST_COLOR_AXIS;
        }
        if ( wc<HST_NCOLORS )
        {
            if ( !gdk_color_parse(color_name,h->colors+wc) )
            {
                g_printerr("Cannot get color %s\n",color_name);
            }
            else
                hst1d_set_color(h,wc,h->colors+wc);
        }
    } while(A);
    va_end(var_args);
}

static void
hst1d_draw(GtkHst1D *hst)
{
    g_return_if_fail (hst != NULL);
    g_return_if_fail (GTK_IS_HST1D (hst));

    if (!GTK_WIDGET_REALIZED (GTK_WIDGET(hst)))
        return;

    if (hst->colors_changed)
    {
        hst1d_set_style_colors(hst);
        hst1d_set_curve_colors(hst);
        hst->colors_changed=FALSE;
    }

    hst->gc = gdk_gc_new (GTK_WIDGET(hst)->window);
    gtk_paint_flat_box (GTK_WIDGET (hst)->style, hst->pm,
                        GTK_STATE_NORMAL,
                        GTK_SHADOW_NONE,
                        NULL,
                        GTK_WIDGET (hst), "",
                        0, 0,
                        GTK_WIDGET (hst)->allocation.width,
                        GTK_WIDGET (hst)->allocation.height);
    /*
     * Draw frame finds the maximum and minimum values to be
     * histogramed. It certainly deserves a better name...
     */
    GTK_HST1D_GET_CLASS(hst)->draw_frame(hst);
    canvas_set(hst->cnvs,hst);
    GTK_HST1D_GET_CLASS(hst)->draw_pixmap(hst);

    if ( hst->cnvs->legend )
    {
        gdk_draw_layout (hst->pm,
                         hst->gc,
                         hst->cnvs->leg_x,
                         hst->cnvs->leg_y,
                         hst->cnvs->legend
            );
    }

    GTK_HST1D_GET_CLASS(hst)->draw_axis(hst);
    /* Show the statistics */
    if ( gtk_hst1d_stats(hst) )
        GTK_HST1D_GET_CLASS(hst)->draw_stats(hst);

    gdk_draw_pixmap (GTK_WIDGET(hst)->window,
                     hst->gc,
                     hst->pm,
                     0, 0, 0, 0, -1, -1);

    g_object_unref(hst->gc);
}

static void
hst1d_draw_pixmap(GtkHst1D *hst)
{
    int ic,nc;
    GList *lc = hst->cnvs->curve_colors;

    /* draw the list of points */
    nc = g_list_length(hst->curves);

    for (ic=0;ic<nc;ic++)
    {
        GdkPoint *pt = canvas_transform_tuple(hst->cnvs,hst->xpt+ic);
        if ( lc )
        {
            GdkColor *col = (GdkColor *)lc->data;
            gdk_gc_set_foreground (hst->gc,col);
            if (!lc->next)
                lc = g_list_first(lc);
            else
                lc = lc->next;
        }
        else
            gdk_gc_set_foreground (hst->gc,hst->colors+HST_COLOR_FOREGROUND);

        if ( hst->xpt[ic].n>0 )
        {
            if ( hst->attr & HST_FILLED )
            {
                double xmin = hst->cnvs->ymin < 0. ? 0. : hst->cnvs->ymin;
                int    n = hst->xpt[ic].n;
                pt = (GdkPoint *)g_realloc( pt,(n+3)*sizeof(GdkPoint));
                canvas_transform(hst->cnvs,0.,xmin,pt+n);
                pt[n].x = pt[n-1].x;
                n++;
                canvas_transform(hst->cnvs,0.,xmin,pt+n);
                pt[n].x = pt[0].x;
                n++;
                pt[n].x = pt[0].x;
                pt[n].y = pt[0].y;
                n++;
                gdk_draw_polygon (hst->pm, hst->gc, 1, pt,n);
            }
            else
                gdk_draw_lines (hst->pm, hst->gc, pt, hst->xpt[ic].n);
        }
        g_free(pt);
        hst_function_clear_tuple(hst->xpt+ic);
    }
    g_free(hst->xpt);
}

/* TODO */
static
void hst1d_draw_stats(GtkHst1D *h)
{
    static char line[256];
    PangoLayout    *layout;
    PangoRectangle  fsize;
    HstStats        *stats;
    Canvas        *cnvs;
    HstHistogram  *hst;
    HstFunction   *func;
    if (!HST_IS_HISTOGRAM(h->func))
        return;

    
    hst = HST_HISTOGRAM(h->func);
    func = HST_FUNCTION(hst);
    stats = hst->stat;
    /*
     * Check if we have a fixed range
     */
    if ( h->xmin != func->xmin || h->xmax !=func->xmax )
    {
        int i, ib1, ib2,n;
        double x, step;
        // update the range statistics
        hst_stats_reset(h->cnvs->stats);
        ib1 = hst_function_find_bin(func, h->cnvs->xmin);
        if (ib1<0)
            ib1=0;
        if (ib1<=func->nbin)
            ib1 = func->nbin-1;
        
        ib2 = hst_function_find_bin(func, h->cnvs->xmax);
        if (ib2<0)
            ib2 = 0;
        if (ib2>=func->nbin)
            ib2 = func->nbin-1;
        
        step = func->step;
        x = h->cnvs->xmin + 0.5*step;
        n = (ib2-ib1)+1;
        for (i=ib1;i<=ib2;i++, x+=step)
        {
            hst_stats_add(h->cnvs->stats, x,0.,hst->y[i]);
        }
        stats = h->cnvs->stats;
    }
    
    sprintf(line,"<span size=\"smaller\">Entries : % .4e\nMean   : %.4e\nRMS    : %.4e</span>",
            hst_stats_get_nc(stats),
            hst_stats_get_mean(stats,0),
            hst_stats_get_stdev(stats,0)
        );

    layout = gtk_widget_create_pango_layout (GTK_WIDGET(h),NULL);
    pango_layout_set_markup(layout,line,-1);
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    gdk_draw_layout (h->pm,h->gc,h->cnvs->width - fsize.width,h->cnvs->ch,layout);
    g_object_unref(layout);

}

static void
hst1d_draw_frame(GtkHst1D *h)
{
    int ic;
    GList *tmp_list;
    Canvas *cnvs = h->cnvs;

    /*
      Get the maximum, the minimum and the list of points
    */
    cnvs->xmin = h->xmin;
    cnvs->xmax = h->xmax;
    if (!cnvs->fixed)
    {
        cnvs->ymin = G_MAXDOUBLE;
        cnvs->ymax = -G_MAXDOUBLE;
    }
    h->xpt = g_new(Tuple,g_list_length(h->curves));

    for(ic=0,tmp_list=g_list_first(h->curves);tmp_list;ic++)
    {
        HstFunction *f = tmp_list->data;
        h->xpt[ic].n=0;
        h->xpt[ic].P=0;
        hst_function_get_array(f,h->xpt+ic,h->xmin,h->xmax);

        if ( !cnvs->fixed )
        {
            int ip;
            for (ip=0;ip<h->xpt[ic].n;ip++)
            {
                double x = h->xpt[ic].P[ip].x;
                double y = h->xpt[ic].P[ip].y;
                if ( y < cnvs->ymin )
                    cnvs->ymin = y;
                if ( y > cnvs->ymax  )
                    cnvs->ymax = y;

                if ( gtk_hst1d_is_logy(h) )
                    h->xpt[ic].P[ip].y = y > 0. ? log10(y) : 0.;
                if ( gtk_hst1d_is_logx(h) )
                    h->xpt[ic].P[ip].x = x > 0. ? log10(x) : 0.;
            }
        }
        tmp_list = g_list_next (tmp_list);
    }
    if ( !cnvs->fixed )
    {
        double yrange;
        if ( cnvs->ymin==cnvs->ymax )
            cnvs->ymax = cnvs->ymin + 10.;

        yrange = cnvs->ymax - cnvs->ymin;
        if (cnvs->ymin!=0.)
            cnvs->ymin -= 0.1*yrange; //(cnvs->ymin < 0. ? 1.1 : 0.9) * cnvs->ymin;
        
        cnvs->ymax += 0.1*yrange; //(cnvs->ymax < 0. ? 0.9 : 1.1) * cnvs->ymax;
    }
}

static void
hst1d_draw_axis(GtkHst1D *h)
{
    PangoLayout    *layout;
    PangoRectangle  fsize;

    /*------------------------------------------------------------------------
     * Draw the axis
     *------------------------------------------------------------------------*/
    if (h->drawaxis)
    {
        gdk_gc_set_foreground(h->gc,h->colors+HST_COLOR_AXIS);
        /* Draw Y axis   */
        if (h->drawaxis & HST_AXIS_Y_MASK)
        {
            int axl = (h->drawaxis >> 4) & 0xf;
            draw_axis (h,HST_Y_AXIS,axl,gtk_hst1d_is_logy(h));
        }
        /* Draw X axis  */
        if (h->drawaxis & HST_AXIS_X_MASK)
        {
            int axl = h->drawaxis & 0xf;
            draw_axis (h,HST_X_AXIS,axl,gtk_hst1d_is_logx(h));
        }

        if ( GTK_IS_HST2D(h) )
        {
            // Draw the Z axis
            draw_axis (h,HST_Z_AXIS,1,FALSE);
        }
        gdk_gc_set_foreground(h->gc,h->colors+HST_COLOR_FOREGROUND);
    }
    if ( h->evt->button & HST_BUTTON_PRESS )
    {
        double dx = h->evt->x - h->evt->x0;
        double dy = h->evt->y - h->evt->y0;
        sprintf(line,
                "<span size=\"smaller\">X0: %.3g Y0: %.3g X1 %.3g Y1 %.3g DX %.3g DY %.3g dx/dy %.3g</span>",
                h->evt->x0,h->evt->y0,h->evt->x,h->evt->y,dx,dy,dx/dy);
    }
    else
    {
        sprintf(line,"<span size=\"smaller\">X: %.3g Y: %3g</span>",h->evt->x,h->evt->y);
    }

    layout = gtk_widget_create_pango_layout (GTK_WIDGET(h),NULL);
    pango_layout_set_markup(layout,line,-1);
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    gdk_draw_layout (h->pm,h->gc,0,0,layout);
    g_object_unref(layout);
}


void
canvas_set(Canvas *cnvs,GtkHst1D *h)
{
    GtkWidget *w = GTK_WIDGET(h);
    PangoLayout    *layout;
    PangoRectangle  fsize;

    int xaxis_opts = h->drawaxis & HST_AXIS_X_MASK;
    int yaxis_opts = (h->drawaxis & HST_AXIS_Y_MASK)>>4;
    double y_min,y_max;

    /* Get order of magnitude of X axis  */
    get_order_of_magnitude (cnvs->xmax-cnvs->xmin,3., &cnvs->sx, &cnvs->ex);
    if (cnvs->ex != 0.)
    {
        cnvs->sx = pow (10, cnvs->ex);
        sprintf (line, "%.1e", cnvs->sx);
    }
    else
        cnvs->sx = 1.;

    /* Get order of magnitude of Y axis  */
    get_order_of_magnitude (cnvs->ymax-cnvs->ymin, 3., &cnvs->sy, &cnvs->ey);
    if (cnvs->ey != 0.)
    {
        cnvs->sy = pow (10, cnvs->ey);
        sprintf (line, "%.1e", cnvs->sy);
    }
    else
        cnvs->sy = 1.;


    /* Get the character height */
    layout = gtk_widget_create_pango_layout (GTK_WIDGET(h),"X");
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    cnvs->cw = fsize.width;
    cnvs->ch = fsize.height;

    pango_layout_set_markup(layout,"<span size=\"smaller\">X</span>",-1);
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    cnvs->scw = fsize.width;
    cnvs->sch = fsize.height;

    /* Get the origin */
    cnvs->width = w->allocation.width;
    cnvs->height= w->allocation.height;

    /* Witdh of Y axis labels gives the X axis origin   */
    sprintf (line, "%  +0.4g", ceil(cnvs->ymax/cnvs->sy));
    pango_layout_set_markup(layout,line,-1);
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    if ( yaxis_opts & HST_AXIS_LEFT )
        cnvs->ox = fsize.width + 2 * cnvs->cw;
    else
        cnvs->ox = cnvs->cw;

    /* Height of X axis labels sets the Y axis origin   */
    cnvs->oy = (int)(1.5*cnvs->sch);
    if ( xaxis_opts & HST_AXIS_LEFT )
        cnvs->oy = (int) (2.* cnvs->ch + .5);
    else
        cnvs->oy = (int)(1.5*cnvs->ch);


    if ( GTK_IS_HST2D(h) )
    {
        /* Get order of magnitude of Z axis  */
        get_order_of_magnitude (cnvs->zmax-cnvs->zmin, 3., &cnvs->sz, &cnvs->ez);
        if (cnvs->ez != 0.)
        {
            cnvs->sz = pow (10, cnvs->ez);
            sprintf (line, "%.1e", cnvs->sz);
        }
        else
            cnvs->sz = 1.;

        // Width of scale and its axis
        cnvs->cscale.width  = (int)(3.5*cnvs->cw);

        sprintf (line, "%.4g", ceil(cnvs->zmax/cnvs->sz));
        pango_layout_set_markup(layout,line,-1);
        pango_layout_get_pixel_extents (layout, NULL, &fsize);

        cnvs->wz = (int)(cnvs->cscale.width + 3.5*cnvs->cw);
    }
    else
        cnvs->wz = cnvs->ox;


    /*
      Set the space left to draw the axis and the origin and end of
      the axis.
    */

    /* X axis */
    cnvs->wx = cnvs->width  - cnvs->ox - cnvs->wz;
    cnvs->px0 = cnvs->ox;
    cnvs->px1 = cnvs->ox + cnvs->wx;

    /* Y axis */
    if ( xaxis_opts & HST_AXIS_LEFT )
    {
        if ( h->xtitle )
            cnvs->py0 = cnvs->height - (int)(2.5*cnvs->ch);
        else
            cnvs->py0 = cnvs->height - (int)(1.25*cnvs->ch);
    }
    else
        cnvs->py0 = cnvs->height - cnvs->ch;

    cnvs->py1 = cnvs->oy;
    cnvs->wy = (cnvs->py0 - cnvs->py1);

    if ( GTK_IS_HST2D(h) )
    {
        GtkHst2D *h2d = GTK_HST2D(h);
        if (h2d->squared)
        {
            if ( cnvs->wx>cnvs->wy )
            {
                cnvs->wx  = cnvs->wy;
                cnvs->px1 = cnvs->px0 + cnvs->wx;
            }
            else
            {
                cnvs->wy  = cnvs->wx;
                cnvs->py0 = cnvs->py1 + cnvs->wy;
            }
        }
    }

    if ( cnvs->legend )
    {
        PangoRectangle  lsiz;
        pango_layout_get_pixel_extents (cnvs->legend, NULL, &lsiz);
        cnvs->wx   -= lsiz.width;
        cnvs->leg_x = cnvs->ox+cnvs->wx + (int)(LEGEND_XTRA*lsiz.width/2.);
        cnvs->leg_y = cnvs->py1 + (cnvs->wy - lsiz.height)/2;
    }

    if ( GTK_IS_HST2D(h) )
    {
        cnvs->cscale.height = cnvs->wy;
        cnvs->cscale.x      = cnvs->px1+h->cnvs->cw;
        cnvs->cscale.y      = cnvs->py1;

        cnvs->wz  = cnvs->wy;
        cnvs->pz0 = cnvs->py0;
        cnvs->pz1 = cnvs->py1;
        cnvs->oz  = cnvs->cscale.x + cnvs->cscale.width;
        cnvs->zstp= cnvs->wz/(cnvs->zmax - cnvs->zmin);
    }


    /* Get the transf. factor */
    cnvs->logx=FALSE;
    if ( gtk_hst1d_is_logx(h) )
    {
        double l1 = cnvs->xmax <=0 ? 1. : log10(cnvs->xmax);
        double l2 = cnvs->xmin <=0 ? 0. : log10(cnvs->xmin);
        cnvs->xstp = cnvs->wx/( rint(l1-l2)+1. );
        cnvs->logx=TRUE;
        cnvs->lxmin=l2;
    }
    else
        cnvs->xstp = cnvs->wx/(cnvs->xmax - cnvs->xmin);

    y_max = cnvs->ymax;
    y_min = cnvs->ymin;
    cnvs->logy=FALSE;
    if ( gtk_hst1d_is_logy(h) )
    {
        double l1 = y_max <=0 ? 1. : log10(y_max);
        double l2 = y_min <=0 ? 0. : log10(y_min);
        cnvs->ystp = (cnvs->py0 - cnvs->py1) / (rint(l1-l2) +1.);
        cnvs->logy=TRUE;
        cnvs->lymin=l2;
    }
    else
    {
        cnvs->ystp = (cnvs->py0 - cnvs->py1) / (y_max - y_min);
    }

    g_object_unref(layout);
}

GdkPoint *
canvas_transform_tuple(Canvas *cnvs, Tuple *t)
{
    int i,n = t->n;
    GdkPoint *P = g_new(GdkPoint,n);
    for (i=0;i<n;i++)
        canvas_transform(cnvs,t->P[i].x,t->P[i].y,P+i);

    return P;
}

inline void
canvas_transform(Canvas *cnvs,
                 double x, double y, GdkPoint *pt)
{
    double xmin = cnvs->logx ? cnvs->lxmin : cnvs->xmin;
    double ymin = cnvs->logy ? cnvs->lymin : cnvs->ymin;



    pt->x = (int)(cnvs->px0 + (x-xmin)*cnvs->xstp );
    pt->y = (int)( cnvs->py0 - cnvs->ystp*(y-ymin) );
    if ( pt->y > cnvs->py0 )
        pt->y = cnvs->py0;
    if ( pt->y < cnvs->py1 )
        pt->y = cnvs->py1;
}

inline void
canvas_transform_inv(Canvas *cnvs,int ix, int iy, double *x, double *y)
{
    double xmin = cnvs->logx ? cnvs->lxmin : cnvs->xmin;
    double ymin = cnvs->logy ? cnvs->lymin : cnvs->ymin;

    *x = ( ix - cnvs->px0 )/cnvs->xstp + xmin;
    *y = ymin -( iy - cnvs->py0 )/cnvs->ystp;
}




static int axis_height;
#define GET_Y_POINT(y_val,y_min,factor) (axis_height-((y_val)-(y_min))*(factor))
#define GET_Y_VALUE(y_point,y_min,factor) ((y_min)+(axis_height-(y_point))/(factor))
#define GET_X_POINT(x_val,x_min,factor) (((x_val)-(x_min))*(factor))
#define GET_X_VALUE(x_point,x_min,factor) ((x_min)+(x_point)/(factor))


void
draw_axis (GtkHst1D *hst,GtkHstAxisType axtype, int axopts,int logx)
{
    int Ndiv[2] = { 10, 5 }, nBins[2];
    float Wmn[2], Wmx[2], Bw[2] = { 1., 1. };
    float x, x1, xx, val,x0;
    int iyl, ch;
    int point[4];
    int x_axis = (axtype == HST_X_AXIS);
    int ox, oy, aw;
    double x_min, x_max, fAxis,scale, x2nd;
    const char *title;
    PangoLayout    *layout;
    PangoRectangle  fsize;
    double eeps;

    layout = gtk_widget_create_pango_layout (GTK_WIDGET(hst),NULL);


    if (axtype == HST_X_AXIS)
    {
        title = hst->xtitle;
        ox = hst->cnvs->px0;
        if ( axopts & HST_AXIS_LEFT)
        {
            oy = hst->cnvs->py0;
        }
        else if (axopts & HST_AXIS_MIDDLE)
        {
            oy = hst->cnvs->py0 - hst->cnvs->wy / 2;
        }
        else
        {
            oy = hst->cnvs->py1;
        }
        aw = hst->cnvs->wx;
        x_min = hst->cnvs->xmin;
        x_max = hst->cnvs->xmax;
        fAxis = hst->cnvs->xstp;
        scale = logx ? 1. : hst->cnvs->sx;
    }
    else if ( axtype == HST_Y_AXIS )
    {
        title = hst->ytitle;
        oy = hst->cnvs->py0;
        if ( axopts & HST_AXIS_LEFT)
        {
            ox = hst->cnvs->px0;
        }
        else if (axopts & HST_AXIS_MIDDLE)
        {
            ox = hst->cnvs->px0 + hst->cnvs->wx / 2;
        }
        else
        {
            ox = hst->cnvs->px1;
        }
        aw = hst->cnvs->wy;
        if ( GTK_IS_HST2D(hst) )
        {
            g_object_get(G_OBJECT(hst),"hymin",&x_min,NULL);
            g_object_get(G_OBJECT(hst),"hymax",&x_max,NULL);
        }
        else
        {
            x_min = hst->cnvs->ymin;
            x_max = hst->cnvs->ymax;
        }
        fAxis = hst->cnvs->ystp;
        scale = logx ? 1. : hst->cnvs->sy;
    }
    else
    {
        /* Z axis: color scale */
        title = NULL;
        oy = hst->cnvs->py0;
        ox = hst->cnvs->oz;
        aw = hst->cnvs->wz;
        fAxis = hst->cnvs->zstp;
        scale = hst->cnvs->sz;
        x_min = hst->cnvs->zmin;
        x_max = hst->cnvs->zmax;
    }

    if ( logx )
    {
        x_min = (x_min>0. ? log10(x_min) : 0.);
        x_max = rint( (x_max>0. ? log10(x_max) : 0.) + 0.5 );
    }
    eeps = fabs(x_max-x_min)*eps;

    if (title==0)
        title=" ";

    ch = hst->cnvs->ch;
    axis_height = oy;

//    Ndiv[0] = logx ? 5 : 10;
    getbins (x_min, x_max, Ndiv[0], Wmn, Wmx, nBins, Bw);
    if (fabs(x_min - Wmn[0]) > eeps)
    {
        Wmn[0] -= Bw[0];
        nBins[0]++;
    }
    if (fabs(Wmx[0] - x_max) > eeps)
    {
        Wmx[0] += Bw[0];
        nBins[0]++;
    }
    x2nd = logx ? pow(10.,Bw[0]) : Bw[0];
    getbins (0, x2nd, Ndiv[1], Wmn + 1, Wmx + 1, nBins + 1, Bw + 1);


    gdk_draw_line (hst->pm, hst->gc, ox, oy, x_axis ? ox + aw : ox, x_axis ? oy : oy - aw);

    ch = hst->cnvs->ch;
    iyl = (axtype = HST_X_AXIS ? oy + (int) (1.25 * ch) + 2 : ch / 2);

    for (x = Wmn[0]; x <= Wmx[0]; x += Bw[0])
    {
        val = x / scale;
        if (fabs (val) < 1.e-5)
            val = 0.;
        xx  = val;
        x0  = x;
        if (logx)
        {
            /*
              val = val <=0. ? x_min : log10(val);
              x0  = x0  <=0. ? x_min : log10(x);
            */
            xx = pow(10.,xx);
        }
        sprintf (line, "%.3g", xx);
        point[0] = x_axis ? ox + GET_X_POINT (x0, x_min, fAxis) : ox;
        point[1] = x_axis ? oy : GET_Y_POINT (x0, x_min, fAxis);
        point[2] = x_axis ? point[0] : ox + 5.;
        point[3] = x_axis ? oy - 5 : point[1];

        if ( x0>=x_min && x0<=x_max )
        {
            gdk_draw_line (hst->pm, hst->gc, point[0], point[1], point[2], point[3]);
            pango_layout_set_text(layout,line,-1);
            pango_layout_get_pixel_extents (layout, NULL, &fsize);
            gdk_draw_layout (hst->pm, hst->gc,
                             point[0] - (x_axis ? fsize.width / 2 : fsize.width + 4),
                             point[1] + (x_axis ? 0.2*fsize.height : -fsize.height/2),
                             layout);
        }
        for (x1 = Wmn[1] + Bw[1]; x1 < Wmx[1] + eeps ; x1 += Bw[1])
        {
            if (logx)
            {
                val = x + log10(x1);
            }
            else
                val = x+x1;

            if (val<=x_min || val>=x_max)
                continue;

            point[0] = x_axis ? ox + GET_X_POINT (val, x_min, fAxis) : ox;
            point[1] = x_axis ? oy : GET_Y_POINT (val, x_min, fAxis);
            point[2] = x_axis ? point[0] : ox + 2;
            point[3] = x_axis ? oy - 2 : point[1];
            gdk_draw_line (hst->pm, hst->gc, point[0], point[1], point[2], point[3]);
        }
    }

    /* Axis title */
    line[0]=0;
    if (scale!=1.)
    {
        int si = log10 (scale);
        if (title)
            sprintf (line, "<b>%s</b>  x10<sup>%d</sup>", title,si);
        else
            sprintf (line, "x10<sup>%d</sup>", si);
    }
    else
    {
        if (title)
            sprintf (line, "<b>%s</b>",title);
    }
    if (line[0])
    {
        pango_layout_set_markup(layout,line,-1);
        pango_layout_get_pixel_extents (layout, NULL, &fsize);
        gdk_draw_layout (hst->pm,hst->gc,
                         x_axis ? ox + aw - fsize.width : ox - ch,
                         x_axis ? GTK_WIDGET(hst)->allocation.height-fsize.height : oy - aw-fsize.height,
                         layout
            );
    }
    g_object_unref(layout);
}



static gboolean
hst1d_button_press(GtkHst1D *hst, GdkEventButton *ev, gpointer *d)
{
    double xd,x = ev->x;
    double yd,y = ev->y;

    if ( ev->button==2 )
    {
        hst->xmin = hst->func->xmin;
        hst->xmax = hst->func->xmax;
        g_object_set(hst,"vmin",0.,NULL);
        g_object_set(hst,"vmax",0.,NULL);
        gtk_widget_queue_draw(GTK_WIDGET(hst));
    }
    else if ( ev->button == 3 )
    {
        GtkWidget *menu = hst1d_create_menu(hst);
        gtk_menu_popup (GTK_MENU (menu), 0, 0, 0, 0, 0, 0);
    }
    else if ( ev->button!=1 )
        canvas_transform_inv(hst->cnvs,x,y,&xd,&yd);

    hst->evt->button |= HST_BUTTON_PRESS;
    hst->evt->x0 = hst->evt->x;
    hst->evt->y0 = hst->evt->y;

    return FALSE;
}

static gboolean
hst1d_button_release(GtkHst1D *hst, GdkEventButton *ev, gpointer *data)
{
    double xd,x = ev->x;
    double yd,y = ev->y;

    hst->evt->button |= HST_BUTTON_RELEASE;
    hst->evt->button = ~HST_BUTTON_PRESS;


    if ( ev->button!=1 )
        return FALSE;

    canvas_transform_inv(hst->cnvs,x,y,&xd,&yd);
    if ( ev->state & GDK_CONTROL_MASK )
    {
        hst->xmin = hst->evt->x0 < xd ? hst->evt->x0 : xd;
        hst->xmax = hst->evt->x0 > xd ? hst->evt->x0 : xd;
    }
    if ( ev->state & GDK_SHIFT_MASK )
    {
        g_object_set(hst,"vmin",hst->evt->y0 < yd ? hst->evt->y0 : yd,NULL);
        g_object_set(hst,"vmax",hst->evt->y0 > yd ? hst->evt->y0 : yd,NULL);
    }
    if ( ev->state & (GDK_CONTROL_MASK|GDK_SHIFT_MASK) )
        gtk_widget_queue_draw(GTK_WIDGET(hst));

    return FALSE;
}

static gboolean
hst1d_motion_notify(GtkHst1D *hst, GdkEventMotion *ev, gpointer *data)
{
    double xd,x = ev->x;
    double yd,y = ev->y;

    if ( ev->state & GDK_CONTROL_MASK )
        return FALSE;

    canvas_transform_inv(hst->cnvs,x,y,&xd,&yd);
    hst->evt->x = xd;
    hst->evt->y = yd;
    gtk_widget_queue_draw(GTK_WIDGET(hst));
    return FALSE;
}

static gboolean
hst1d_key_press(GtkHst1D *hst, GdkEventKey *ev, gpointer *data)
{
    printf("Key %d %x\n",ev->keyval,ev->keyval);
    return TRUE;
}

enum {
    CFOREG,
    CBACKG,
    CAXIS,
    CXAXIS_LEFT,
    CXAXIS_RIGHT,
    CXAXIS_MIDDLE,
    CXAXIS_NONE,
    CYAXIS_LEFT,
    CYAXIS_RIGHT,
    CYAXIS_MIDDLE,
    CYAXIS_NONE,
    CFILL,
    CLOGY,
    CLOGX,
    CSTATS,
    C2D_BW,
    C2D_WB,
    C2D_RED,
    C2D_GREEN,
    C2D_BLUE,
    C2D_THERMAL,
    LAST_C
};

struct item_data
{
    int item;
    GtkHst1D *hst;
};
struct item_data itmData[LAST_C];

static void     hst1d_create_colorsel (struct item_data *);


static void
on_item_activate (GtkWidget * w, struct item_data *p)
{
    switch (p->item)
    {
    case CFOREG:
    case CBACKG:
    case CAXIS:
        hst1d_create_colorsel(p);
        break;
    case CXAXIS_LEFT:
        gtk_hst1d_draw_x_axis(p->hst,HST_AXIS_LEFT);
        break;
    case CXAXIS_RIGHT:
        gtk_hst1d_draw_x_axis(p->hst,HST_AXIS_RIGHT);
        break;
    case CXAXIS_MIDDLE:
        gtk_hst1d_draw_x_axis(p->hst,HST_AXIS_MIDDLE);
        break;
    case CXAXIS_NONE:
        gtk_hst1d_draw_x_axis(p->hst,HST_AXIS_NONE);
        break;
    case CYAXIS_LEFT:
        gtk_hst1d_draw_y_axis(p->hst,HST_AXIS_LEFT);
        break;
    case CYAXIS_RIGHT:
        gtk_hst1d_draw_y_axis(p->hst,HST_AXIS_RIGHT);
        break;
    case CYAXIS_MIDDLE:
        gtk_hst1d_draw_y_axis(p->hst,HST_AXIS_MIDDLE);
        break;
    case CYAXIS_NONE:
        gtk_hst1d_draw_y_axis(p->hst,HST_AXIS_NONE);
        break;
    case CFILL:
        gtk_hst1d_filled(p->hst,gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)));
        break;
    case CLOGX:
        gtk_hst1d_logx(p->hst,gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)));
        break;
    case CLOGY:
        gtk_hst1d_logy(p->hst,gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)));
        break;
    case CSTATS:
        gtk_hst1d_show_stats(p->hst,gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w)));
        break;
    case C2D_BW:
    case C2D_WB:
    case C2D_RED:
    case C2D_GREEN:
    case C2D_BLUE:
    case C2D_THERMAL:
        break;
    default:
        break;
    }
}

static GtkWidget *
add_item(GtkWidget *menu, const char *name)
{
    GtkWidget *item;
    item = gtk_menu_item_new_with_mnemonic(name);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    return item;
}

static GtkWidget *
add_submenu( GtkWidget *menu )
{
    GtkWidget *submenu = gtk_menu_new ();
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu), submenu);
    return submenu;
}

static GtkWidget *
add_action(GtkWidget *menu, const char *name, int code, gboolean check, gboolean val)
{
    GtkWidget *item;
    if (check)
    {
        item = gtk_check_menu_item_new_with_mnemonic (name);
        gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (item), val);
    }
    else
        item = gtk_menu_item_new_with_mnemonic (name);

    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    g_signal_connect ((gpointer) item, "activate",
                      G_CALLBACK (on_item_activate),
                      itmData+code);
    return item;
}

static GtkWidget *
add_separator( GtkWidget *menu)
{
    GtkWidget *item = gtk_separator_menu_item_new ();
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_widget_set_sensitive (item, FALSE);
    return item;
}

static void
on_btn_print_browse(GtkWidget *w, GtkEntry *entry)
{
    GtkWidget *fsel;
    fsel = gtk_file_selection_new ("Output jpeg");
    gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (fsel));
    gtk_file_selection_complete (GTK_FILE_SELECTION (fsel), "*.jp*");
    if ( gtk_dialog_run (GTK_DIALOG (fsel)) == GTK_RESPONSE_OK )
    {
        const char *fnam = gtk_file_selection_get_filename(GTK_FILE_SELECTION (fsel));
        if (fnam)
            gtk_entry_set_text(entry,fnam);
    }
    gtk_widget_hide(fsel);
    gtk_widget_destroy(fsel);
}

static void
on_hst_print(GtkWidget *w, GtkHst1D *h)
{
    GtkWidget *dlg;
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *label;
    GtkWidget *entry;
    GtkWidget *btn;
    GtkWidget *quality;

    dlg = gtk_dialog_new_with_buttons(
        "Save Histogram as jpg or png",
        NULL,
        GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
        GTK_STOCK_OK,
        GTK_RESPONSE_ACCEPT,
        GTK_STOCK_CANCEL,
        GTK_RESPONSE_REJECT,
        NULL
        );
    gtk_window_set_title(GTK_WINDOW(dlg),"Save Histogram as jpg or png");
    gtk_window_set_type_hint(GTK_WINDOW(dlg),GDK_WINDOW_TYPE_HINT_DIALOG);

    vbox = gtk_vbox_new(FALSE,4);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox),vbox, FALSE,TRUE,0);

    label = gtk_label_new("File name");
    gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
    gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);

    hbox = gtk_hbox_new(FALSE,0);
    gtk_box_pack_start(GTK_BOX(vbox),hbox,TRUE,TRUE,0);

    entry = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox),entry,TRUE,TRUE,0);

    btn = gtk_button_new_from_stock("gtk-open");
    g_signal_connect (G_OBJECT(btn), "clicked",
                      G_CALLBACK (on_btn_print_browse),
                      entry);
    gtk_box_pack_start(GTK_BOX(hbox),btn,FALSE,FALSE,0);

    label = gtk_label_new("Quality");
    gtk_misc_set_alignment(GTK_MISC(label),0,0.5);
    gtk_box_pack_start(GTK_BOX(vbox),label,FALSE,FALSE,0);

    quality = gtk_hscale_new(
        GTK_ADJUSTMENT(gtk_adjustment_new (100, 0, 100, 1, 10, 0))
        );
    gtk_box_pack_start(GTK_BOX(vbox),quality,TRUE,TRUE,0);

    gtk_widget_show_all(dlg);

    if ( gtk_dialog_run( GTK_DIALOG(dlg) )==GTK_RESPONSE_ACCEPT )
    {
        GdkPixbuf *pxb=NULL;
        GError *err = NULL;
        const char *fnam = gtk_entry_get_text(GTK_ENTRY(entry));
        printf("The file is %s \n",fnam);

        g_object_get(G_OBJECT(h),"pixmap",&pxb,NULL);
        if (!pxb)
        {
            g_error("Could not get the histogram pixmap\n");
        }
        else
        {
            if (!gdk_pixbuf_save (pxb,fnam, "jpeg",
                                  &err,
                                  "quality", "100",
                                  NULL))
            {
                g_error("Failed creating the jpeg: %s",err->message);
                g_error_free(err);
            }
            g_object_unref(G_OBJECT(pxb));
        }
    }

    gtk_widget_hide(dlg);
    gtk_widget_destroy(dlg);
}

static GtkWidget *
hst1d_create_menu( GtkHst1D *h)
{
    GtkWidget *menu;
    GtkWidget *item;
    GtkWidget *item_menu;
    GtkWidget *subitem;
    int i;

    for (i = 0; i < LAST_C; i++)
    {
        itmData[i].item = i;
        itmData[i].hst = h;
    }


    menu = gtk_menu_new ();

    /* Print */
    item      = add_item(menu,"Print...");
    g_signal_connect (G_OBJECT(item), "activate",
                      G_CALLBACK (on_hst_print),
                      h
        );

    item = add_separator(menu);


    /* Colors*/
    item      = add_item(menu,"colors");
    item_menu = add_submenu(item);
    subitem   = add_action(item_menu,"Foreground...",CFOREG,FALSE,FALSE);
    subitem   = add_action(item_menu,"Background...",CBACKG,FALSE,FALSE);
    subitem   = add_action(item_menu,"Axis...",CAXIS,FALSE,FALSE);

    item = add_separator(menu);

    /* Axis */
    item      = add_item(menu,"X axis...");
    item_menu = add_submenu(item);
    subitem   = add_action(item_menu,"down"  ,CXAXIS_LEFT,FALSE,FALSE);
    subitem   = add_action(item_menu,"middle",CXAXIS_MIDDLE,FALSE,FALSE);
    subitem   = add_action(item_menu,"top"   ,CXAXIS_RIGHT,FALSE,FALSE);
    subitem   = add_action(item_menu,"none"  ,CXAXIS_NONE,FALSE,FALSE);

    /* Y axis options */
    item      = add_item(menu,"Y axis...");
    item_menu = add_submenu(item);
    subitem   = add_action(item_menu,"left"  ,CYAXIS_LEFT,FALSE,FALSE);
    subitem   = add_action(item_menu,"middle",CYAXIS_MIDDLE,FALSE,FALSE);
    subitem   = add_action(item_menu,"right" ,CYAXIS_RIGHT,FALSE,FALSE);
    subitem   = add_action(item_menu,"none"  ,CYAXIS_NONE,FALSE,FALSE);

    item = add_separator(menu);

    /* Hst1d attributes */
    item = add_action(menu,"Show stats",CSTATS,TRUE,gtk_hst1d_stats(h));
    item = add_action(menu,"filled",CFILL,TRUE,gtk_hst1d_is_filled(h));
    item = add_action(menu,"log(x)",CLOGX,TRUE,gtk_hst1d_is_logx(h));
    item = add_action(menu,"log(y)",CLOGY,TRUE,gtk_hst1d_is_logy(h));

    /* Add any extra menu */
    if ( GTK_HST1D_GET_CLASS(h)->xtra_menu !=NULL )
        GTK_HST1D_GET_CLASS(h)->xtra_menu(h,menu);

    return menu;
}

/*-----------------------------------------------------------------
 * This is the part that controls the color selection in the popup
 * menu of the histogram
 *-----------------------------------------------------------------*/

struct _color_dlg_
{
    GtkHst1D *hst;
    GtkColorSelectionDialog *cs;
    int which_color;
};

static void
on_color_selected (GtkButton * button, gpointer user_data)
{
    struct _color_dlg_ *dt = (struct _color_dlg_ *) user_data;
    GtkColorSelection *cs = GTK_COLOR_SELECTION (dt->cs->colorsel);
    GdkColor cc;
    gtk_color_selection_get_current_color (cs, &cc);
    gtk_widget_destroy (GTK_WIDGET (dt->cs));

    hst1d_set_color(dt->hst,dt->which_color-CFOREG,&cc);
    gtk_widget_queue_draw(GTK_WIDGET(dt->hst));
}

static void
on_color_cancel (GtkButton * button, gpointer user_data)
{
    struct _color_dlg_ *dt = (struct _color_dlg_ *) user_data;
    gtk_widget_destroy (GTK_WIDGET (dt->cs));
}

static void
hst1d_create_colorsel (struct item_data *itmD)
{
    GtkWidget *dlg;
    GtkColorSelection *colorsel;
    GtkWidget *ok_btn;
    GtkWidget *cancel_btn;
    GdkColor  *color=0;
    char dname[50];
    static struct _color_dlg_ dlg_data;

    switch (itmD->item)
    {
    case CFOREG:
        color = itmD->hst->colors;
        sprintf (dname, "Choose foreground color");
        break;
    case CBACKG:
        color = itmD->hst->colors+1;
        sprintf (dname, "Choose background color");
        break;
    case CAXIS:
        color = itmD->hst->colors+2;
        sprintf (dname, "Choose axis color");
        break;
    default:
        sprintf (dname, "Choose   Unknown   color");
        break;
    }
    dlg = gtk_color_selection_dialog_new (dname);
    colorsel = GTK_COLOR_SELECTION( GTK_COLOR_SELECTION_DIALOG(dlg)->colorsel);
    gtk_color_selection_set_previous_color (colorsel, color);
    gtk_color_selection_set_current_color (colorsel, color);
    gtk_color_selection_set_has_palette (colorsel, TRUE);

    gtk_container_set_border_width (GTK_CONTAINER (colorsel), 10);
    dlg_data.hst = itmD->hst;
    dlg_data.cs =  GTK_COLOR_SELECTION_DIALOG (dlg);
    dlg_data.which_color = itmD->item;

    ok_btn     = GTK_COLOR_SELECTION_DIALOG (dlg)->ok_button;
    cancel_btn = GTK_COLOR_SELECTION_DIALOG (dlg)->cancel_button;

    gtk_signal_connect (GTK_OBJECT (ok_btn),
                        "clicked",
                        GTK_SIGNAL_FUNC (on_color_selected),
                        &dlg_data);
    gtk_signal_connect (GTK_OBJECT (cancel_btn),
                        "clicked",
                        GTK_SIGNAL_FUNC (on_color_cancel),
                        &dlg_data);

    gtk_widget_show_all( dlg );
}

static void
hst1d_range_changed(HstFunction *f,GtkHst1D *h)
{
    g_return_if_fail( GTK_IS_HST1D(h) );
    h->xmin = f->xmin;
    h->xmax = f->xmax;
    gtk_widget_queue_draw(GTK_WIDGET(h));
}
