#include <math.h>
#include "stats.h"

/* here are local prototypes */
static void stats_get_property (GObject * object,
                                guint property_id,
                                GValue * value,
                                GParamSpec * pspec);
static void stats_init (HstStats * o);
static void stats_class_init (HstStatsClass * c);

enum
{
    PROP_0,
    PROP_NC,
    PROP_X_MEAN,
    PROP_X_RMS,
    PROP_X_STDEV,
    PROP_X_MIN,
    PROP_X_MAX,
    PROP_Y_MEAN,
    PROP_Y_RMS,
    PROP_Y_STDEV,
    PROP_Y_MIN,
    PROP_Y_MAX,
    PROP_CORR
};

/* pointer to the class of our parent */
static GObjectClass *parent_class = NULL;



GType
hst_stats_get_type (void)
{
    static GType type = 0;

    if (type == 0)
    {
        static const GTypeInfo info = {
            sizeof (HstStatsClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) stats_class_init,
            (GClassFinalizeFunc) NULL,
            NULL /* class_data */ ,
            sizeof (HstStats),
            0 /* n_preallocs */ ,
            (GInstanceInitFunc) stats_init,
            NULL
        };

        type = g_type_register_static (G_TYPE_OBJECT,
                                       "HstStats", &info,
                                       (GTypeFlags) 0);
    }
    return type;
}


static void
stats_init (HstStats * o)
{
    hst_stats_reset(o);
}

static void
stats_class_init (HstStatsClass * c)
{
    GObjectClass *g_object_class = (GObjectClass *) c;

    parent_class = g_type_class_ref (G_TYPE_OBJECT);

    g_object_class->get_property = stats_get_property;

    g_object_class_install_property (
        g_object_class,
        PROP_NC,
        g_param_spec_double ("nc",
                             "nc",
                             "Number of entries",
                             -G_MAXDOUBLE ,
                             G_MAXDOUBLE,
                             0.0 ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );
    g_object_class_install_property (
        g_object_class,
        PROP_X_MEAN,
        g_param_spec_double ("mean_x" ,
                             "X mean" ,
                             "Mean value of X" ,
                             -G_MAXDOUBLE ,
                             G_MAXDOUBLE,
                             0.0 ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );
    g_object_class_install_property (
        g_object_class,
        PROP_Y_MEAN,
        g_param_spec_double ("mean_y" ,
                             "Y mean" ,
                             "Mean value of Y" ,
                             -G_MAXDOUBLE ,
                             G_MAXDOUBLE,
                             0.0 ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );

    g_object_class_install_property (
        g_object_class,
        PROP_X_RMS,
        g_param_spec_double ("rms_x" /* name */ ,
                             "X rms" /* nick */ ,
                             "RMS of X" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );
    g_object_class_install_property (
        g_object_class,
        PROP_Y_RMS,
        g_param_spec_double ("rms_y" /* name */ ,
                             "Y rms" /* nick */ ,
                             "RMS of Y" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );


    g_object_class_install_property (
        g_object_class,
        PROP_X_STDEV,
        g_param_spec_double ("stdev_x" /* name */ ,
                             "X stdev" /* nick */ ,
                             "Standard deviation of X" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );
    g_object_class_install_property (
        g_object_class,
        PROP_Y_STDEV,
        g_param_spec_double ("stdev_y" /* name */ ,
                             "Y stdev" /* nick */ ,
                             "Standard deviation of Y" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );

    g_object_class_install_property (
        g_object_class,
        PROP_CORR,
        g_param_spec_double ("corr" /* name */ ,
                             "XY correlation" /* nick */ ,
                             "X and Y correlation" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );

    g_object_class_install_property (
        g_object_class,
        PROP_X_MIN,
        g_param_spec_double ("xmin" /* name */ ,
                             "xmin" /* nick */ ,
                             "X minimum value" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );
    g_object_class_install_property (
        g_object_class,
        PROP_X_MAX,
        g_param_spec_double ("xmax" /* name */ ,
                             "xmax" /* nick */ ,
                             "X maximum value" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );

    g_object_class_install_property (
        g_object_class,
        PROP_Y_MIN,
        g_param_spec_double ("ymin" /* name */ ,
                             "ymin" /* nick */ ,
                             "Y minimum value" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );
    g_object_class_install_property (
        g_object_class,
        PROP_Y_MAX,
        g_param_spec_double ("ymax" /* name */ ,
                             "ymax" /* nick */ ,
                             "Y maximum value" /* blurb */ ,
                             -G_MAXDOUBLE /* minimum */ ,
                             G_MAXDOUBLE /* maximum */ ,
                             0.0 /* default_value */ ,
                             (GParamFlags) (G_PARAM_READABLE)
            )
        );


}

static void
stats_get_property (GObject * object,
                    guint property_id,
                    GValue * VAL,
                    GParamSpec * pspec)
{
    HstStats *self;
    self = HST_STATS (object);

    switch (property_id)
    {
    case PROP_NC:
        g_value_set_double (VAL, self->nc);
        break;
    case PROP_X_MEAN:
        g_value_set_double (VAL,hst_stats_get_mean(self,0));
        break;
    case PROP_X_RMS:
        g_value_set_double (VAL, hst_stats_get_rms(self,0));
        break;
    case PROP_X_STDEV:
        g_value_set_double (VAL,hst_stats_get_stdev(self,0));
        break;
    case PROP_X_MIN:
        g_value_set_double( VAL, hst_stats_get_min(self,0));
        break;
    case PROP_X_MAX:
        g_value_set_double (VAL, hst_stats_get_max(self,0));
        break;
    case PROP_Y_MEAN:
        g_value_set_double (VAL,hst_stats_get_mean(self,1));
        break;
    case PROP_Y_RMS:
        g_value_set_double (VAL, hst_stats_get_rms(self,1));
        break;
    case PROP_Y_STDEV:
        g_value_set_double (VAL,hst_stats_get_stdev(self,1));
        break;
    case PROP_Y_MIN:
        g_value_set_double (VAL,hst_stats_get_min(self,1));
        break;
    case PROP_Y_MAX:
        g_value_set_double (VAL,hst_stats_get_max(self,1));
        break;

    case PROP_CORR:
        g_value_set_double(VAL,hst_stats_get_correlation(self));
        break;
        
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        if (&VAL || &pspec)
            break;
        break;
    }

}

/**
 * hst_stats_new:
 *
 * This is a convinience function to a create a HstStats object.
 *
 * Return value: a #HstStats object.
 */
HstStats *
hst_stats_new()
{
    HstStats *s = (HstStats *)g_object_new(hst_stats_get_type(), NULL);
    return s;
}

/**
 * hst_stats_get_nc:
 * @self: a #HstStats object
 * 
 * Return value: the number of entries
 */
gdouble
hst_stats_get_nc (HstStats *self)
{
    g_return_val_if_fail ( self != NULL ,0);
    return self->nc;
}

/**
 * hst_stats_get_mean:
 * @self: a #HstStats object
 * @w: specifies which of the 2 variables is to be considered.
 *     (=0) is X and (=1) is Y.
 * 
 * Return value: the mean value of the selected variable
 */
gdouble
hst_stats_get_mean (HstStats * self, int w)
{
    g_return_val_if_fail (self != NULL, 0.);
    g_return_val_if_fail ( (w>=0&&w<2), 0. );

    return self->nc > 0. ? self->s[w] / self->nc : 0.;
}

/**
 * hst_stats_get_rms:
 * @self: a #HstStats object
 * @w: specifies which of the 2 variables is to be considered.
 *     (=0) is X and (=1) is Y.
 * 
 * Return value: the RMS of the selected variable
 */
gdouble
hst_stats_get_rms (HstStats * self, int w)
{
    double val,mean;
    g_return_val_if_fail (self != NULL, 0.);
    g_return_val_if_fail ( (w>=0&&w<2), 0. );

    if (self->nc < 2.)
        val = 0.;
    else
    {
        mean = hst_stats_get_mean (self,w);
        val = (self->s[2+w] / self->nc - mean * mean);
    }

    return val;
}

/**
 * hst_stats_get_stdev:
 * @self: a #HstStats object
 * @w: specifies which of the 2 variables is to be considered.
 *     (=0) is X and (=1) is Y.
 * 
 * Return value: the standard deviation of the selected variable
 */
gdouble 
hst_stats_get_stdev (HstStats * self, int w)
{
    g_return_val_if_fail (self != NULL, 0.);
    g_return_val_if_fail ( (w>=0&&w<2), 0. );

    return sqrt( hst_stats_get_rms(self,w) );
}

/**
 * hst_stats_get_min:
 * @self: a #HstStats object
 * @w: specifies which of the 2 variables is to be considered.
 *     (=0) is X and (=1) is Y.
 * 
 * Return value: the minimum value of the selected variable
 */
gdouble 
hst_stats_get_min (HstStats * self, int w)
{
    g_return_val_if_fail (self != NULL, 0.);
    g_return_val_if_fail ( (w>=0&&w<2), 0. );

    return self->xmin[w];
}

/**
 * hst_stats_get_max:
 * @self: a #HstStats object
 * @w: specifies which of the 2 variables is to be considered.
 *     (=0) is X and (=1) is Y.
 * 
 * Return value: the maximum value of the selected variable
 */
gdouble 
hst_stats_get_max (HstStats * self, int w)
{
    g_return_val_if_fail (self != NULL, 0.);
    g_return_val_if_fail ( (w>=0&&w<2), 0. );

    return self->xmax[w];
}

/**
 * hst_stats_get_correlation:
 * @self: a #HstStats object
 * 
 * Return value: the correlation between X and Y
 */
gdouble 
hst_stats_get_correlation (HstStats *self)
{
    double val=0.;
    g_return_val_if_fail (self != NULL, 0.);

    if ( self->nc > 0.)
    {
        gdouble x = hst_stats_get_mean(self,0);
        gdouble y = hst_stats_get_mean(self,1);
        gdouble sxy = hst_stats_get_stdev(self,0)*hst_stats_get_stdev(self,1);
        val = (sxy ==0. ? 0. : (self->s[4]/self->nc - x*y)/sxy);
    }
    return val;
}


/**
 * hst_stats_add:
 * @self: a #HstStats object
 * @x: value of X
 * @y: value of Y
 * @w: the weight of this point
 * 
 * Adds a new point with a given point
 */
void 
hst_stats_add (HstStats * self,double x, double y, double w)
{
    int i;
    double xv;
    g_return_if_fail (self != NULL);

    self->nc += w;
    for ( i=0; i<2; i++ )
    {
        xv = i ? y : x;
        self->s[i] += w * xv;
        self->s[2+i] += w * xv * xv;

        if (xv > self->xmax[i])
            self->xmax[i] = xv;
        if (xv < self->xmin[i])
            self->xmin[i] =xv;
    }
    self->s[4] += x * y;
}

/**
 * hst_stats_reset:
 * @self: a #HstStats object
 * 
 * Resets the statistics. 
 */
void 
hst_stats_reset (HstStats * self)
{
    int i;
    g_return_if_fail (self != NULL);
    self->nc = 0.0;
    for (i=0;i<5;i++)
        self->s[i]=0.;
    for (i=0;i<2;i++)
    {
        self->xmin[i] =  G_MAXDOUBLE;
        self->xmax[i] = -G_MAXDOUBLE;
    }
}
