#include <stdio.h>
#include "tracer.h"

enum {
    DESTROY,
    LAST_SIGNAL
};
enum {
    PROP_0
};

static void       tracer_base_class_init     (HstTracerClass *class);
static void       tracer_base_class_finalize (HstTracerClass *class);
static void       tracer_class_init          (HstTracerClass *klass);
static void       tracer_init                (HstTracer      *object,
                                              HstTracerClass *klass);
static void       tracer_dispose             (GObject        *object);
static void       tracer_finalize            (GObject        *object);
static void       tracer_array               (HstFunction *f,
                                              Tuple *,
                                              double xmin, double xmax);
static void       tracer_fill                (HstHistogram *hst,
                                              double x, double w);
static int        tracer_resize              (HstFunction *f,
                                              int width, int height);
static gpointer    parent_class = NULL;
/* static guint       object_signals[LAST_SIGNAL] = { 0 }; */

GType hst_tracer_get_type (void)
{
    static GType object_type = 0;

    if (!object_type)
    {
        static const GTypeInfo object_info = {
            sizeof (HstTracerClass),
            (GBaseInitFunc) tracer_base_class_init,
            (GBaseFinalizeFunc) tracer_base_class_finalize,
            (GClassInitFunc) tracer_class_init,
            NULL,               /* class_finalize */
            NULL,               /* class_data */
            sizeof (HstTracer),
            16,         /* n_preallocs */
            (GInstanceInitFunc) tracer_init,
            NULL,               /* value_table */
        };

        object_type = g_type_register_static (HST_TYPE_HISTOGRAM,
                                              "HstTracer", &object_info, 0);
    }

    return object_type;
}

static void 
tracer_base_class_init (HstTracerClass *class)
{
    /* reset instance specifc methods that don't get inherited */
}

static void 
tracer_base_class_finalize (HstTracerClass *class)
{
}

static void
tracer_class_init (HstTracerClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (class);
    HstHistogramClass *hclass = HST_HISTOGRAM_CLASS(class);
    HstFunctionClass  *fclass = HST_FUNCTION_CLASS(class);
    parent_class = g_type_class_ref (HST_TYPE_HISTOGRAM);

    gobject_class->dispose = tracer_dispose;
    gobject_class->finalize = tracer_finalize;

    fclass->resize      = tracer_resize;
    fclass->array       = tracer_array;

    hclass->fill        = tracer_fill;
}

static void
tracer_init( HstTracer *t, HstTracerClass *klass)
{
    g_return_if_fail( HST_IS_TRACER(t) );

    t->index=0;
    t->npoints=0;
}

/**
 * hst_tracer_new:
 * @n: number of points in the tracer
 *
 * This is a convinience function to greate a tracer with a buffer of
 * size @n
 *
 * Return value: A new #HstTracer
 */
HstTracer *
hst_tracer_new( int n )
{
    HstTracer    *t;
    double    xmax=n;
    g_return_val_if_fail ( n>0, NULL );

    t = g_object_new (HST_TYPE_TRACER, "nx",n,"xmin",0.,"xmax",xmax,"ptr",NULL,NULL);

    return t;
}

static void
tracer_dispose(GObject *object)
{
    G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
tracer_finalize (GObject *object)
{
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
tracer_fill(HstHistogram *h, double x, double w)
{
    HstTracer *T;
    HstFunction *f;
    g_return_if_fail( HST_IS_TRACER(h) );

    T = HST_TRACER(h);
    f = HST_FUNCTION(h);
    if ( T->index >= f->nbin )
        T->index=0;
    if ( T->npoints<f->nbin )
        T->npoints++;


    h->y[T->index++] = x;
    hst_stats_add(h->stat,x,0.,w);

    if ( x > h->vmax )
        h->vmax = x;
    if ( x < h->vmin )
        h->vmin = x;


}

static void
tracer_array(HstFunction *f, Tuple *t, double xmin, double xmax)
{
    HstTracer *T;
    HstHistogram *h;
    int i,j,ofs,npts;

    t->n = 0;
    t->P = 0;

    g_return_if_fail( HST_IS_TRACER(f) );

    T = HST_TRACER(f);
    h = HST_HISTOGRAM(f);
    npts = T->npoints;
    if ( npts>f->nbin )
        npts = f->nbin;

    t->n = npts;
    t->P = g_new(PointList,t->n);
    ofs = T->index - T->npoints;
    if (ofs<0)
        ofs += f->nbin;

    for (i=0,j=ofs;i<npts;i++,j++) {
        t->P[i].x = i;
        if (j>=f->nbin)
            j=0;
        t->P[i].y = h->
            y[j];
    }
}

static int
tracer_resize(HstFunction *f,int width, int height)
{
    HstTracer *T;
    HstHistogram *h;
    g_return_val_if_fail( HST_IS_TRACER(f),0);
    T = HST_TRACER(f);
    h = HST_HISTOGRAM(f);
    if ( f->nbin != width ) {
        f->nbin = width;
        f->xmax = width;
        h->y = (double *)g_realloc( h->y, f->nbin * sizeof(double) );
        if ( T->index>=f->nbin )
            T->index=0;
        if ( T->npoints>f->nbin )
            T->npoints = f->nbin;
        
        return 1;
    }
    return 0;
}
